"""Bathymetry class and utilities."""

import random
from typing import Optional, Sequence, Tuple, Union

from absl import logging

import matplotlib
import numpy as np
import utm

import inductiva


class Bathymetry:
    """Represents a bathymetric profile.
    
    A bathymetric profile defines the depth of the sea bottom as a function of
    space, here described in Cartesian coordinates (x, y).
    
    Here, a bathymetry is represented with:
    - a 1D array of depths, in meters, measured at arbitrary points in space.
      Positive depths are below the water level.
    - two 1D arrays representing the x and y coordinates of the points where the
      depths are defined, in meters.
    """

    def __init__(
        self,
        depths: np.ndarray,
        x: np.ndarray,
        y: np.ndarray,
    ):
        """Initializes a `Bathymetry` object.
        
        Args:
            depths: A 1D array with the depths, in meters.
            x: A 1D array with the x coordinates of the points where depths are
              defined, in meters.
            y: Same as `x`, but for the y coordinates.
        """
        self.depths = depths
        self.x = x
        self.y = y

    @classmethod
    def from_bot_file(
        cls,
        bot_file_path: str,
        x_range: Sequence[float],
        y_range: Sequence[float],
    ):
        """Creates a `Bathymetry` object from a bot file.
        
        The depth values are read from a bot file, i.e. a text file with a 2D
        array with the depths, in meters. The first and second dimensions of the
        array in the text file (i.e. rows and columns) correspond to the x and y
        directions, respectively.

        Args:
            text_file_path: Path to the text file.
            x_range: The range of x values, in meters.
            y_range: The range of y values, in meters.
        """

        depths = np.loadtxt(bot_file_path)

        x, y = np.meshgrid(np.linspace(*x_range, depths.shape[0]),
                           np.linspace(*y_range, depths.shape[1]),
                           indexing="ij")

        return cls(depths.flatten(), x.flatten(), y.flatten())

    @classmethod
    def from_ascii_xyz_file(
        cls,
        ascii_xyz_file_path: str,
        remove_offset: bool = True,
    ):
        """Creates a `Bathymetry` object from an ASCII XYZ file.
        
        ASCII XYZ files store bathymetric data in a table where each line
        corresponds to a location (latitude, longitude pair). Several columns
        are available to characterize the depth at each location, namely:
        - longitude, in decimal degrees, e.g. 52.07334567;
        - latitude, in decimal degrees, e.g. 3.06033283;
        - the minum depth, in meters, e.g. 35.81;
        - the maximum depth, in meters, e.g. 35.81;
        - average depth (over a set of measurements), in meters, e.g. 35.81.
        - the standard deviation of the depth (over a set of measurements), in
          meters, e.g. 35.81;

        For more information, see
        https://emodnet.ec.europa.eu/sites/emodnet.ec.europa.eu/files/public/20171127_DTM_exchange_format_specification_v1.6.pdf.

        In this method, we:
        - load the latitude and longitude values from the file, and convert
          them to (x, y) UTM coordinates (see
          https://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system
          for more information).
        - use average depth as the local depth value.

        Args:
            ascii_xyz_file_path: Path to the ASCII XYZ file.
            remove_offset: Whether to remove the offset of the x and y
              coordinates. If `True`, the minimum x and y values are subtracted
              from all x and y values, respectively.
        """

        # Load only the first, second and fifth columns of the file,
        # corresponding to longitude, latitude and average depth, respectively.
        bathymetry_data = np.loadtxt(ascii_xyz_file_path,
                                     usecols=(0, 1, 4),
                                     delimiter=";")

        x_lon = bathymetry_data[:, 0]
        y_lat = bathymetry_data[:, 1]
        depths = bathymetry_data[:, 2]

        x, y, _, _ = utm.from_latlon(y_lat, x_lon)

        if remove_offset:
            x = x - x.min()
            y = y - y.min()

        return cls(depths, x, y)

    @classmethod
    def from_random_depths(
        cls,
        x_range: Sequence[float],
        y_range: Sequence[float],
        x_num: int,
        y_num: int,
        max_depth: float = 10,
        initial_roughness: float = 1,
        roughness_factor: float = 0.5,
        percentile_above_water: float = 20,
    ):
        """Creates a `Bathymetry` object with random depths.

        The depths of the corners of the grid are chosen according to a maximum
        depth value `max_depth` and a percentile of the domain that must be
        above water `percentile_above_water`. The corners on the lower x
        boundary (East) are assumed to be below water (i.e. have 0 < depths <
        `max_depth`). The corners on the upper x boundary (West) are assumed to
        be above water (i.e. have - `max_depth` < depths < 0).
        
        Args:
            x_range: The range of x values, in meters.
            y_range: The range of y values, in meters.
            x_num: Number of grid points in the x direction.
            y_num: Number of grid points in the y direction.
            max_depth: Maximum depth value, in meters.
            initial_roughness: Initial roughness value, in meters. Controls the
              initial range of randomness of the Diamond-Square algorithm.
            roughness_factor: Roughness factor. Must be between 0 and 1.
              Controls the rate at which the range of randomness of the
              Diamond-Square algorithm decreases over iterations.
            percentile_above_water: Percentile of the depths that must be above
              water. Must be between 0 and 100.
        """

        corner_values = [
            random.uniform(0, max_depth),
            random.uniform(0, max_depth),
            random.uniform(-max_depth, 0),
            random.uniform(-max_depth, 0),
        ]

        depths = inductiva.generative.procedural.generate_random_map_level(
            x_num,
            y_num,
            corner_values,
            initial_roughness,
            roughness_factor,
        )

        depths = inductiva.generative.procedural.adjust_map_level(
            depths, percentile_above_water)

        x, y = np.meshgrid(np.linspace(*x_range, x_num),
                           np.linspace(*y_range, y_num),
                           indexing="ij")

        return cls(depths.flatten(), x.flatten(), y.flatten())

    def to_bot_file(
        self,
        bot_file_path: str,
    ):
        """Writes the bathymetry to a bot file.

        The depth values are interpolated to a regular grid and written to a bot
        file. The grid size is determined automatically from the range and the
        number of unique x and y values.

        Args:
            text_file_path: Path to the text file.
        """

        x_resolution = self.x_ptp() / self.x_uniques().size
        y_resolution = self.y_ptp() / self.y_uniques().size

        depths_grid, _ = self._interpolate_to_uniform_grid(
            x_resolution, y_resolution)

        np.savetxt(bot_file_path, depths_grid)

    @property
    def x_range(self) -> Tuple[float]:
        """Returns the range of x values."""

        return (np.min(self.x), np.max(self.x))

    @property
    def y_range(self) -> Tuple[float]:
        """Returns the range of y values."""

        return (np.min(self.y), np.max(self.y))

    def x_ptp(self) -> float:
        """Returns the peak-to-peak range (max - min) of x values."""

        return np.ptp(self.x)

    def y_ptp(self) -> float:
        """Returns the peak-to-peak range (max - min) of y values."""

        return np.ptp(self.y)

    def x_uniques(self, sort: bool = False) -> np.ndarray:
        """Returns the unique x values.
        
        Args:
            sort: Whether to sort the unique values.
        """
        x_uniques = np.unique(self.x)
        if sort:
            x_uniques = np.sort(x_uniques)
        return x_uniques

    def y_uniques(self, sort: bool = False) -> np.ndarray:
        """Returns the unique y values.
        
        Args:
            sort: Whether to sort the unique values.
        """
        y_uniques = np.unique(self.y)
        if sort:
            y_uniques = np.sort(y_uniques)
        return y_uniques

    def crop(self,
             x_range: Sequence[float],
             y_range: Sequence[float],
             remove_offset=True):
        """Crops the bathymetry to a given range of x and y values.
        
        Args:
            x_range: The range of x values, in meters.
            y_range: The range of y values, in meters.    
        """
        x_min, x_max = x_range
        y_min, y_max = y_range

        mask = (self.x >= x_min) & (self.x <= x_max) & \
               (self.y >= y_min) & (self.y <= y_max)

        if np.sum(mask) == 0:
            raise ValueError(
                "The bathymetry cannot be cropped because no points are "
                "defined in the given coordinate ranges.")

        x = self.x[mask]
        y = self.y[mask]
        depths = self.depths[mask]

        if remove_offset:
            x = x - x.min()
            y = y - y.min()

        return Bathymetry(
            depths=depths,
            x=x,
            y=y,
        )

    def is_uniform_grid(self) -> bool:
        """Determines whether the bathymetry is defined on a uniform grid."""

        x_uniques_diffs = np.diff(self.x_uniques(sort=True))
        y_uniques_diffs = np.diff(self.y_uniques(sort=True))

        return (np.unique(x_uniques_diffs.round(decimals=2)).size == 1 and
                np.unique(y_uniques_diffs.round(decimals=2)).size == 1)

    def plot(
        self,
        cmap: Optional[str] = None,
        clim: Optional[Tuple[float]] = None,
        path: Optional[str] = None,
        x_resolution: float = 10,
        y_resolution: float = 10,
        threshold_distance: float = 20,
    ) -> Union[matplotlib.axes.Axes, None]:
        """Plots the bathymetry.

        The bathymetry is represented as a 2D map of depths, with the x and y
        coordinates of the points where the depths are defined in the axes.

        The bathymetry data is plotted with a color plot on a grid with uniform
        spacing in the x and y directions. The spacing in the x and y directions
        is controlled by the `x_resolution` and `y_resolution` arguments.
        
        The data is interpolated from the points where the bathymetry is defined
        to the uniform grid using linear interpolation.

        Points on the uniform grid at a distance larger than a threshold
        distance `threshold_distance` from points where the bathymetry is
        defined are omitted.
    
        The plot is produced with matplotlib.

        Args:
            cmap: Colormap to use. Defaults to the matplotlib default colormap.
            clim: Range of depth values to represent in colors. If `None`, the
              range is determined automatically from the minumum and maximum
              depth. Depth values below or above this range are represented with
              the minimum or maximum colors, respectively.
            path: Path to save the plot. If `None`, the plot is not saved, and
              the matplotlib `Axes` object is returned instead.
            x_resolution: Resolution, in meters, of the plotting grid in the x
              direction.
            y_resolution: Resolution, in meters, of the plotting grid in the y
              direction.
            threshold_distance: Threshold distance to filter out points on the
              uniform grid that are far from points where the bathymetry is
              defined.
        """

        logging.info("Plotting the bathymetry on a uniform grid...")
        depths_grid, _ = self._interpolate_to_uniform_grid(
            x_resolution,
            y_resolution,
            threshold_distance=threshold_distance,
            nullable=True,
        )

        x_size = self.x_ptp() / x_resolution
        y_size = self.y_ptp() / y_resolution

        if x_size > 1000 and y_size > 1000:
            logging.warning(
                "The plotting grid is large. It may take a while to plot.")

        # Plot the bathymetry.
        extent = (
            self.x_range[0],
            self.x_range[1],
            self.y_range[0],
            self.y_range[1],
        )

        fig = matplotlib.pyplot.figure()
        ax = fig.add_subplot()

        im = ax.imshow(
            depths_grid.transpose(),
            cmap=cmap,
            clim=clim,
            origin="lower",
            extent=extent,
        )

        ax.set(
            aspect="equal",
            xlim=self.x_range,
            ylim=self.y_range,
            xlabel="$x$ [m]",
            ylabel="$y$ [m]",
        )

        fig.colorbar(im, ax=ax, label="Depth [m]")

        if path is not None:
            fig.savefig(path)
            matplotlib.pyplot.close(fig)

        else:
            return ax

    def to_uniform_grid(
        self,
        x_resolution: float = 2,
        y_resolution: float = 2,
        fill_value: Optional[Union[float, str]] = None,
    ):
        """Converts the bathymetry to a uniform grid.

        The bathymetry is interpolated to a grid with uniform resolution (i.e.
        spacing) in the x and y directions. Linear interpolation is used to
        interpolate the bathymetry to the grid.
        
        Grid points for which no interpolation is possible may be filled with
        a constant or with the nearest depth value.

        Args:
            x_resolution: Resolution, in meters, of the grid in the x direction.
            y_resolution: Resolution, in meters, of the grid in the y direction.
            fill_value: Value to fill the grid points for which no interpolation
              is possible. If "nearest", the nearest depth value is used.
        """

        depths_grid, (x_grid, y_grid) = self._interpolate_to_uniform_grid(
            x_resolution, y_resolution, fill_value=fill_value, nullable=False)

        return Bathymetry(depths=depths_grid.flatten(),
                          x=x_grid.flatten(),
                          y=y_grid.flatten())

    def _interpolate_to_uniform_grid(
        self,
        x_resolution: float,
        y_resolution: float,
        threshold_distance: Optional[float] = None,
        fill_value: Optional[Union[float, str]] = None,
        nullable: bool = False,
    ):
        """Interpolates the bathymetry to a uniform grid.     

        Args:
            x_resolution: Resolution, in meters, of the grid in the x direction.
            y_resolution: Resolution, in meters, of the grid in the y direction.
            threshold_distance: Threshold distance to filter out points on the
              uniform grid that are far from points where the bathymetry is
              defined.
            fill_value: Value to fill the grid points for which no interpolation
              is possible. If "nearest", the nearest depth value is used.
            nullable: Whether to allow the bathymetry to be undefined in some
              grid points. If `False`, an error is raised if the bathymetry is
              undefined in one or more grid points.
        
        Returns:
            depths_grid: A 2D array with the depths on the uniform grid.
            (x_grid, y_grid): The x and y coordinates of the points where the
              depths are defined, respectively.
        """
        # Determine grid size based on ranges and resolution.
        x_size = int(self.x_ptp() / x_resolution)
        y_size = int(self.y_ptp() / y_resolution)

        logging.info(
            "Interpolating the bathymetry to a uniform grid...\n"
            "- grid resolution %f x %f (m x m) \n"
            "- grid size %d x %d", x_resolution, y_resolution, x_size, y_size)

        # Create uniform grid for interpolation.
        (x_grid, y_grid) = inductiva.utils.grids.get_meshgrid(
            x_range=self.x_range,
            y_range=self.y_range,
            x_num=x_size,
            y_num=y_size,
        )

        depths_grid = inductiva.utils.interpolation.interpolate_to_coordinates(
            coordinates=(self.x, self.y),
            values=self.depths,
            interpolation_coordinates=(x_grid, y_grid),
            method="linear",
            threshold_distance=threshold_distance)

        nan_mask = np.isnan(depths_grid)

        if fill_value is not None:
            if fill_value == "nearest":
                nearest_depths_grid = \
                    inductiva.utils.interpolation.interpolate_to_coordinates(
                        coordinates=(self.x, self.y),
                        values=self.depths,
                        interpolation_coordinates=(x_grid, y_grid),
                        method="nearest")

                depths_grid[nan_mask] = nearest_depths_grid[nan_mask]
            else:
                depths_grid[nan_mask] = fill_value

        nan_mask = np.isnan(depths_grid)

        if not nullable and np.sum(nan_mask) > 0:
            raise ValueError(
                "The bathymetry cannot be interpolated to a uniform grid "
                "because depths are not defined in one or more grid points.")

        return depths_grid, (x_grid, y_grid)



"""Utils related to template files."""

import os
import io
from typing import Dict, List, Union
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from inductiva.utils.files import find_path_to_package

TEMPLATES_PATH = find_path_to_package("templates")


def replace_params_in_template(
    template_path: str,
    params: Dict,
    output_file: Union[str, io.StringIO],
    remove_template: bool = False,
) -> None:
    """Replaces parameters in a template file."""

    template_path = Path(template_path)

    template_dir = template_path.parent
    template_filename = template_path.name

    environment = Environment(loader=FileSystemLoader(template_dir))
    template = environment.get_template(template_filename)
    stream = template.stream(**params)
    stream.dump(output_file)

    if remove_template:
        os.remove(template_path)


def batch_replace_params_in_template(
    templates_dir: str,
    template_filenames: List[str],
    params: Dict,
    output_filename_paths: List[str],
    remove_templates: bool = False,
) -> None:
    """Replaces parameters in a set of template files.

    For some simulators, more than one file needs to be changed.
    Moreover, some parameters are altered in different input files.
    To simplify, we can do a batch change for the same params dict.

    Args:
        templates_dir: Directory with the template files.
        template_filename_paths: List containing all template files to
            be changed.
        params: Dictionary of params that are inserted in the template
            files.
        output_filename_paths: List containing the output files in regard of
            the template files.
    """

    for index, template_filename in enumerate(template_filenames):
        template_path = os.path.join(templates_dir, template_filename)

        replace_params_in_template(
            template_path=template_path,
            params=params,
            output_file=output_filename_paths[index],
            remove_template=remove_templates,
        )



"""SWASH module of the API."""
from typing import Optional

from inductiva.simulation import Simulator
from inductiva import types, tasks, resources


class SWASH(Simulator):
    """Class to invoke a generic SWASH simulation on the API."""

    @property
    def api_method_name(self) -> str:
        return "sw.swash.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        machine_group: Optional[resources.MachineGroup] = None,
        n_cores: int = 1,
        run_async: bool = False,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            n_cores: Number of MPI cores to use for the simulation.
            sim_config_filename: Name of the simulation configuration file.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           machine_group=machine_group,
                           input_filename=sim_config_filename,
                           n_cores=n_cores,
                           run_async=run_async)



"""Methods to interact with the tasks submitted to the API."""
import json
import datetime
import pandas as pd
import numpy as np
from typing import Dict, List, Optional, Union, Sequence

import inductiva
from inductiva import api
from inductiva.client import ApiClient, ApiException
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.client import models


def _fetch_tasks_from_api(
    status: Optional[Union[str, models.TaskStatusCode]] = None,
    page=1,
    per_page=10,
) -> List[Dict]:
    """Get information about a user's tasks on the API.

    Tags can be filtered by a status. Results are paginated indexed from 1.
    """
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        query_params = {
            "page": page,
            "per_page": per_page,
        }

        if status is not None:
            query_params["status"] = models.TaskStatusCode(status)

        try:
            # Get User Tasks
            resp = api_instance.get_user_tasks(
                query_params=query_params,
                skip_deserialization=True,  # avoid deserializing to model,
                # leave as dict which we'll later serialize to our own
                # dataclasses
            ).response

            response_body = json.loads(resp.data.decode("utf-8"))

            return [{**task} for task in response_body]

        except ApiException as e:
            raise e


def _list_of_tasks_to_str(tasks: Sequence["inductiva.tasks.Task"]) -> str:
    columns = [
        "ID", "Simulator", "Status", "Submitted", "Started", "Duration",
        "VM Type"
    ]
    rows = []

    for task in tasks:
        info = task.get_info()
        # e.g., get "openfoam" from "fvm.openfoam.run_simulation"
        simulator = info["method_name"].split(".")[-2]
        row = [
            task.id,
            simulator,
            task.get_status(),
            info.get("input_submit_time", None),
            info.get("start_time", None),
            task.get_execution_time(),
            task.get_machine_type(),
        ]
        rows.append(row)

    df = pd.DataFrame(rows, columns=columns)

    # replace None with np.nan so that pandas can format them as "n/a"
    # by passing na_rep="n/a" to to_string()
    df.fillna(np.nan, inplace=True)

    def datetime_formatter(dt: str) -> str:
        return datetime.datetime.fromisoformat(dt).strftime("%d %b, %H:%M:%S")

    def seconds_formatter(secs: float) -> str:
        return (f"{int(secs // 3600)}h "
                f"{int((secs % 3600) // 60)}m "
                f"{int(secs % 60)}s")

    formatters = {
        "Submitted": datetime_formatter,
        "Started": datetime_formatter,
        "Duration": seconds_formatter,
    }

    # column width is 15 for all columns except the ones
    # we override below
    col_space = {col: 15 for col in columns}
    col_space["Submitted"] = 20
    col_space["Started"] = 20
    col_space["Status"] = 20
    col_space["VM Type"] = 18

    return df.to_string(
        index=False,
        na_rep="n/a",
        formatters=formatters,
        col_space=col_space,
    )


# pylint: disable=redefined-builtin
def list(num_tasks,
         status: Optional[Union[str, models.TaskStatusCode]] = None) -> None:
    # pylint: disable=line-too-long
    """List the last N tasks of a user.

    This function lists info about the last N tasks (with respect to submission
    time) of a user to stdout, sorted by submission time with the
    most recent first.
    A status can be specified to filter to get only tasks with that status, in
    which case the last N tasks with that status will be listed.
    The number of tasks can be less than N if the aren't enough tasks that
    match the specified criteria.

    Example usage:
        # list the last 5 tasks that were successful
        inductiva.tasks.list(5, status="success")
                         ID       Simulator               Status            Submitted              Started        Duration            VM Type
        1691150776862178362        openfoam              success     04 Aug, 12:06:17     04 Aug, 12:06:18       0h 1m 53s      c2-standard-8
        1691149904961476240        openfoam              success     04 Aug, 11:51:46     04 Aug, 11:51:46       0h 1m 28s      c2-standard-8
        1691081881158823776        openfoam              success     03 Aug, 16:58:02     03 Aug, 16:58:02       0h 1m 20s     n2-standard-32
        1691081409916619414        openfoam              success     03 Aug, 16:50:11     03 Aug, 16:50:11       0h 1m 20s     n2-standard-32
        1691080520213617518        openfoam              success     03 Aug, 16:35:21     03 Aug, 16:35:21       0h 1m 23s     n2-standard-32

    Args:
        num_tasks: The number of most recent tasks with respect to submission
            time to list. If filtering criteria (currently status is available)
            is specified, most recent N tasks that match that criteria will be
            listed. The actual number of tasks may be less if there
            aren't enough tasks available.
        status: The status of the tasks to list. If None, tasks with any status
            will be listed.
    """
    # pylint: enable=line-too-long
    status = models.TaskStatusCode(status) if status is not None else None
    tasks = get(num_tasks, status=status)
    tasks_str = _list_of_tasks_to_str(tasks)
    print(tasks_str)


def get(
    num_tasks,
    status: Optional[Union[str, models.TaskStatusCode]] = None
) -> List["inductiva.tasks.Task"]:
    """Get the last N tasks of a user.

    This function fetches info about the last N tasks (with respect to
    submission time) of a user to stdout, sorted by submission time with the
    most recent first.
    A status can be specified to filter to get only tasks with that status, in
    which case the last N tasks with that status will be listed.
    The number of tasks can be less than N if the aren't enough tasks that match
    the specified criteria.

    Similar to the inductiva.task.list() function, but instead of printing
    to stdout, returns a list of Task objects which can be used to perform
    further operations on those tasks (e.g., download outputs,
    kill unfinished tasks, etc.).

    Example usage:
        # get the last 5 tasks that haven't started yet and kill them
        tasks = inductiva.tasks.get(5, status="submitted")
        for task in tasks:
            task.kill()

    Args:
        num_tasks: The number of most recent tasks with respect to submission
            time to fetch. If filtering criteria (currently status is available)
            is specified, most recent N tasks that match that criteria will be
            listed. The actual number of tasks may be less if there
            aren't enough tasks available.
        status: The status of the tasks to get. If None, tasks with any status
            will be returned.

    Returns:
        List of Task objects.
    """
    status = models.TaskStatusCode(status) if status is not None else None

    raw_tasks_info = _fetch_tasks_from_api(status, page=1, per_page=num_tasks)
    tasks = [
        inductiva.tasks.Task.from_api_info(info) for info in raw_tasks_info
    ]

    return tasks



"""Utils to create the geometric case."""

from typing import List, Optional, Tuple

import gmsh
import json

from inductiva.structures import holes, plates


def get_boundary_ids(entity_gmsh: int) -> List[int]:
    """Gets the IDs of the plate or hole boundaries.

    Args:
        entity_gmsh (int): The Gmsh entity ID representing either a plate or a
          hole's OpenCASCADE CAD representation.

    Returns:
        List[int]: A list of boundary IDs corresponding to the entity's
          boundaries.
    """
    boundaries = gmsh.model.getBoundary([(2, entity_gmsh)])

    return [[boundary_id[1]][0] for boundary_id in boundaries]


class GeometricCase:
    """Geometric case.

    The geometric case is characterized by a plate and a set of holes.

    Attributes:
        plate (RectangularPlate): Rectangular plate object.
        holes_list (List[Hole]): The holes objects.
    """

    def __init__(self,
                 plate: plates.RectangularPlate,
                 holes_list: Optional[List[holes.Hole]] = None) -> None:
        """Initializes a GeometricCase object."""
        self.plate = plate
        self.holes_list = holes_list

    def write_to_json(self, json_path: str) -> None:
        """Writes the geometric case to JSON file.

        Args:
            json_path (str): The JSON file path.
        """

        # Gemeometric case dictionary
        plate_dict = {"plate": self.plate.to_dict()}
        holes_dict = {"holes": [hole.to_dict() for hole in self.holes_list]}
        geom_case_dictionary = {**plate_dict, **holes_dict}

        # Write JSON file
        with open(json_path, "w", encoding="utf-8") as write_file:
            json.dump(geom_case_dictionary, write_file, indent=4)

    def _get_holes_mesh_params(self) -> Tuple[List[float], List[float]]:
        """Gets the mesh generation parameters for all the hole.

        Metrics:
          - mesh_offset (float): Represents an offset for the curves of the
            holes, defining a region around the boundaries. Within this region,
            we have the ability to control the mesh elements size.
          - predefined_element_size (float): Represents the predefined element
            size, defined as 1/4 of the perimeter.

        Returns:
            Tuple[List[float], List[float]]: A tuple containing two lists:
            - List of mesh offsets for each hole.
            - List of predefined element mesh sizes for each hole.
        """
        holes_mesh_offset = []
        holes_predefined_element_size = []

        for hole in self.holes_list:
            mesh_offset, predefined_element_size = hole.get_hole_mesh_params()

            holes_mesh_offset.append(mesh_offset)
            holes_predefined_element_size.append(predefined_element_size)

        return holes_mesh_offset, holes_predefined_element_size

    def _holes_to_occ_and_get_boundary_ids(
            self) -> Tuple[List[int], List[List[int]]]:
        """Converts list of holes to OpenCASCADE CAD, gets boundary IDs.

        Returns:
            Tuple[List[int], List[List[int]]]: A tuple containing two lists:
            - List of the Gmsh entity ID representing the hole's OpenCASCADE
            CAD representation for each hole.
            - List of lists, where each inner list contains the boundary IDs
            corresponding to the hole's boundaries for each hole.
        """

        holes_gmsh = []
        holes_boundary_ids = []

        for hole in self.holes_list:
            hole_gmsh = hole.to_occ()
            gmsh.model.occ.synchronize()

            hole_boundary_ids = get_boundary_ids(hole_gmsh)

            holes_gmsh.append(hole_gmsh)
            holes_boundary_ids.append(hole_boundary_ids)

        return holes_gmsh, holes_boundary_ids

    def palte_with_holes_to_occ_and_get_boundary_ids(
            self) -> Tuple[List[int], List[List[int]]]:
        """Converts palte with holes to OpenCASCADE CAD, gets boundary IDs.

        The process of generating the plate with holes is divided into 3 steps:

        1. Converts plate object to OpenCASCADE CAD representation and gets the
        IDs of the boundaries
        2. Converts holes objects to OpenCASCADE CAD representation and gets the
        IDs of the boundaries
        3. Removes holes from the plate in the OpenCASCADE CAD representation

        To remove the holes from the plate, the gmsh.model.occ.cut() function
        will be used.

        Returns:
            Tuple[List[int], List[List[int]]]:
                A tuple containing the following:
                - List of IDs of the plate's boundaries.
                - List of lists, where each inner list contains the IDs of the
                boundaries corresponding to the boundaries of each hole.
        """

        # 1. Converts the plate object to OpenCASCADE CAD representation and
        # gets the IDs of the boundaries
        plate_gmsh = self.plate.to_occ()
        gmsh.model.occ.synchronize()
        plate_boundary_ids = get_boundary_ids(plate_gmsh)

        # 2. Converts the hole objects to OpenCASCADE CAD representation and
        # gets the IDs of the boundaries
        (holes_gmsh,
         holes_boundary_ids) = self._holes_to_occ_and_get_boundary_ids()

        # 3. Removes holes from the plate in the OpenCASCADE CAD representation
        for hole_gmsh in holes_gmsh:
            gmsh.model.occ.cut([(2, plate_gmsh)], [(2, hole_gmsh)])
        gmsh.model.occ.synchronize()

        return plate_boundary_ids, holes_boundary_ids

    def get_mesh_params(self) -> Tuple[float, float, List[float], List[float]]:
        """Gets the mesh generation parameters for the palte with holes.

        Returns:
            Tuple[float, float, List[float], List[float]]:
                A tuple containing the following:
                - Mesh offset for the plate.
                - Predefined element mesh size for the plate.
                - List of mesh offsets for the holes.
                - List of predefined element sizes for the holes.
        """

        (plate_mesh_offset,
         plate_predefined_element_size) = self.plate.get_plate_mesh_params()
        (holes_mesh_offset,
         holes_predefined_element_size) = self._get_holes_mesh_params()

        return (plate_mesh_offset, plate_predefined_element_size,
                holes_mesh_offset, holes_predefined_element_size)



"""Methods related to managing executers and resource pools."""
from typing import Optional
from uuid import UUID

from absl import logging

import inductiva
from inductiva import api
from inductiva.client import ApiClient, ApiException
from inductiva.client.apis.tags.executers_api import ExecutersApi
from inductiva.client.apis.tags.instance_api import InstanceApi
from inductiva.client.model.instance import Instance
from inductiva.client.model.instance_create import InstanceCreate
from inductiva.client.model.instance_group import InstanceGroup


def create_resource_pool() -> UUID:
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = ExecutersApi(client)

        try:
            api_response = api_instance.create_resource_pool()
        except ApiException as e:
            raise e

    return UUID(api_response.body["id"])


def launch_executer(
    name: str,
    machine_type: str,
    executer_type: str,
    spot: bool = True,
    resource_pool_id: Optional[UUID] = None,
    disk_size_gb: int = 30,
) -> None:

    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)

        logging.info("Launching executer %s of type %s", name, executer_type)

        body = InstanceCreate(
            name=name,
            machine_type=machine_type,
            image_name=executer_type,
            spot=spot,
            resource_pool_id=resource_pool_id,
            disk_size_gb=disk_size_gb,
        )

        try:
            api_instance.create_instance(body=body)
        except ApiException as e:
            raise e


def kill_executer(name: str) -> None:
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)

        logging.info("Killing executer %s", name)
        try:
            api_instance.delete_instance(body=Instance(name=name))
        except ApiException as e:
            raise e


def launch_executer_group(
    num_executers: int,
    name: str,
    machine_type: str,
    executer_type: str,
    spot: bool = True,
    resource_pool_id: Optional[UUID] = None,
    disk_size_gb: int = 10,
) -> None:

    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)

        body = InstanceGroup(
            name=name,
            machine_type=machine_type,
            image_name=executer_type,
            spot=spot,
            resource_pool_id=resource_pool_id,
            num_instances=num_executers,
            disk_size_gb=disk_size_gb,
        )

        try:
            api_instance.create_instance_group(body=body)
        except ApiException as e:
            raise e


def kill_executer_group(name: str) -> None:
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)
        try:
            api_instance.delete_instance_group(body=Instance(name=name))
        except ApiException as e:
            raise e



"""Utils to create the mesh."""

from typing import List, Optional

import gmsh

from . import geometry_utils


def add_mesh_field_distance(mesh_field_id: int, curves_list: List[int],
                            num_points: int) -> None:
    """Add a distance-based mesh size field to control element size in Gmsh.

    - Use gmsh.model.mesh.field.add("Distance", mesh_field_id) to add a
    distance-based mesh size field, where mesh_field_id is a unique identifier
    for the field.
    - Set the curves or points where you want to apply the distance-based field
    using gmsh.model.mesh.field.setNumbers(field_id, "CurvesList", curves_list)
    or gmsh.model.mesh.field.setNumbers(field_id, "PointsList", points_list). 
    Adjust the points_list of the field (i.e., how many points to use along
    curves or points) with 
    gmsh.model.mesh.field.setNumber(field_id, "Sampling", num_points).

    Args:
        mesh_field_id (int): A unique identifier for the distance-based
          mesh size field.
        curves_list (List[int]): List of curve IDs where the
          distance-based field will be applied.
        num_points (int): Number of points to use along curves to adjust
          the field size.
    """

    # Add a distance-based mesh size field using the given field ID
    gmsh.model.mesh.field.add("Distance", mesh_field_id)

    # Set the curves where the distance-based field will be applied
    gmsh.model.mesh.field.setNumbers(mesh_field_id, "CurvesList", curves_list)

    # Adjust the points_list of the field to control mesh size along
    # curves
    gmsh.model.mesh.field.setNumber(mesh_field_id, "Sampling", num_points)


def add_mesh_field_threshold(mesh_field_id: int, ref_mesh_field_id: int,
                             size_min: float, size_max: float, dist_min: float,
                             dist_max: float) -> None:
    """Add a threshold-based mesh size field to control element size in Gmsh.

    - Use gmsh.model.mesh.field.add("Threshold", mesh_field_id) to add a 
    threshold-based mesh size field, where mesh_field_id is a unique identifier
    for the field.
    - Use gmsh.model.mesh.field.setNumber(mesh_field_id, "InField",
    ref_mesh_field_id): The "InField" option specifies that this new field will
    use the values from another existing field to compute its own values. In
    this case, it uses the same field with an identifier ref_mesh_field_id as a
    basis for computation. This means that the new field will be based on the
    values of another field with an identifier one less than the current 
    mesh_field_id. The "InField" option is useful for creating hierarchical mesh
    size fields, where one field depends on another, allowing more control over
    mesh refinement strategies.
    - Set the size constraints for the elements using 
    gmsh.model.mesh.field.setNumber(field_id, "SizeMin", size_min) and 
    gmsh.model.mesh.field.setNumber(field_id, "SizeMax", size_max).
    - Set the distance constraints for the elements using 
    gmsh.model.mesh.field.setNumber(field_id, "DistMin", dist_min) and
    gmsh.model.mesh.field.setNumber(field_id, "DistMax", dist_max).

    Args:
        mesh_field_id (int): A unique identifier for the threshold-based
          mesh size field.
        ref_mesh_field_id (int): The identifier of the existing mesh
          size field to base the new field on.
        size_min (float): Minimum size constraint for the elements.
        size_max (float): Maximum size constraint for the elements.
        dist_min (float): Minimum distance constraint for the elements.
        dist_max (float): Maximum distance constraint for the elements.
    """

    # Add a threshold-based mesh size field using the given field ID
    gmsh.model.mesh.field.add("Threshold", mesh_field_id)

    # Set the reference field that the new field will use to compute its
    #  values
    gmsh.model.mesh.field.setNumber(mesh_field_id, "InField", ref_mesh_field_id)

    # Set the size constraints for the elements.
    gmsh.model.mesh.field.setNumber(mesh_field_id, "SizeMin", size_min)
    gmsh.model.mesh.field.setNumber(mesh_field_id, "SizeMax", size_max)

    # Set the distance constraints for the elements
    gmsh.model.mesh.field.setNumber(mesh_field_id, "DistMin", dist_min)
    gmsh.model.mesh.field.setNumber(mesh_field_id, "DistMax", dist_max)


class GmshMesh:
    """Gmsh mesh.

    Attributes:
        geometry: A geometry_utils.GeometricCase object.
        global_refinement_factor (float): The refinement factor for global
          refinement of the mesh. A higher value results in a finer mesh
          overall, increasing the number of elements in the entire mesh, and
          leading to a more detailed representation of the geometry. Use this
          factor when you want to globally refine the mesh uniformly, without 
          specific local focus.
        local_refinement_factor (float): The refinement factor for local
          refinement of the mesh. This factor controls the local refinement
          level of the mesh and is typically used for refining specific regions
          or features of the mesh. A higher value for this factor indicates a
          finer mesh in the regions of interest, providing more detailed
          resolution around certain features. Use this factor when you want to
          focus on refining specific areas while keeping the rest of the mesh
          less refined.
    """

    def __init__(self,
                 geometry: geometry_utils.GeometricCase,
                 global_refinement_factor: Optional[float] = 1.0,
                 local_refinement_factor: Optional[float] = 10.0) -> None:
        """Initializes a GmshMesh object."""
        self.geometry = geometry
        self.global_refinement_factor = global_refinement_factor
        self.local_refinement_factor = local_refinement_factor

    def _create_mesh_with_gmsh(self) -> None:
        """Creates the mesh with Gmsh.

        To generate the mesh using Gmsh, we utilize mesh size fields: the 
        "Distance" and "Threshold" fields. Use add_mesh_field_distance to add
        the distance mesh field and add_mesh_field_threshold for the threshold
        mesh field in Gmsh.

        To control the element size using distance and threshold mesh size
        fields in Gmsh, we need to follow these steps:

        1. Define the distance-based mesh size field

        2. Define the threshold-based mesh size field

        3. Combine the distance and threshold-based fields:
            - To apply both distance and threshold fields, you can use the field 
            IDs of the previously defined fields and combine them using 
            mathematical operations (e.g., "Min" or "Max").
            - Use gmsh.model.mesh.field.add("Min", combined_mesh_field_id) to
            add a combined field that uses the minimum element size from the 
            distance and threshold fields.
            - Use gmsh.model.mesh.field.setNumbers(combined_mesh_field_id, 
            "FieldsList", list_mesh_field_id) to set the list of mesh fields 
            that will be combined in the newly created field with the 
            identifier combined_mesh_field_id.

        4. Set the mesh field as the background mesh field:
            - Use gmsh.model.mesh.field.setAsBackgroundMesh(
            combined_mesh_field_id) to set the combined field as the
            background mesh field, which will be used during mesh generation.

        5. Generate the mesh:
            - Finally, use gmsh.model.mesh.generate(dim=2) to generate the 2D
            mesh based on the defined mesh size fields and other meshing
            options.

        We can define both the global mesh size and the local mesh size (around
        the holes).
        """

        # Initialize the Gmsh API
        gmsh.initialize()

        # Disable Gmsh terminal output
        gmsh.option.setNumber("General.Terminal", 0)

        # Add a new model and set it as the current model
        gmsh.model.add("model")

        # Generate the plate with holes and get the boundary IDs
        (plate_curves_id, holes_curve_id
        ) = self.geometry.palte_with_holes_to_occ_and_get_boundary_ids()

        # Get mesh parameters
        (plate_mesh_offset, plate_predefined_element_size, holes_mesh_offset,
         holes_predefined_element_size) = self.geometry.get_mesh_params()

        # Add physical markers for the plate and holes
        gmsh.model.addPhysicalGroup(2, [1])

        # Calculate the maximum global mesh size
        global_mesh_size_max = (plate_predefined_element_size /
                                self.global_refinement_factor)

        mesh_field_id = 0
        mesh_field_threshol_id = []

        # Loop through plate boundaries to apply distance and threshold-based
        # mesh fields
        for id_plate_boundary in range(len(plate_curves_id)):

            mesh_field_id += 1
            add_mesh_field_distance(mesh_field_id,
                                    [plate_curves_id[id_plate_boundary]], 100)

            mesh_field_id += 1
            add_mesh_field_threshold(mesh_field_id, mesh_field_id - 1,
                                     global_mesh_size_max, global_mesh_size_max,
                                     0, plate_mesh_offset)
            mesh_field_threshol_id.append(mesh_field_id)

        # Check if local refinement is required
        if self.local_refinement_factor > 0.0:

            # Loop through holes to apply distance and threshold-based mesh
            # fields
            for id_hole in range(len(holes_curve_id)):

                # Calculate the local mesh size
                local_mesh_size = holes_predefined_element_size[
                    id_hole] / self.local_refinement_factor

                # Loop throught curves for each hole
                for hole_curve_id in holes_curve_id[id_hole]:

                    # Set the number of points to use along the hole's curve
                    # (100 for lines, 400 for other types)
                    if gmsh.model.getType(1, hole_curve_id) == "Line":
                        number_points = 100
                    else:
                        number_points = 400

                    mesh_field_id += 1
                    add_mesh_field_distance(mesh_field_id, [hole_curve_id],
                                            number_points)

                    mesh_field_id += 1
                    add_mesh_field_threshold(mesh_field_id, mesh_field_id - 1,
                                             local_mesh_size,
                                             global_mesh_size_max, 0,
                                             holes_mesh_offset[id_hole])
                    mesh_field_threshol_id.append(mesh_field_id)

        # Combine the distance and threshold-based fields
        combined_mesh_field_id = mesh_field_id + 1
        gmsh.model.mesh.field.add("Min", combined_mesh_field_id)
        gmsh.model.mesh.field.setNumbers(combined_mesh_field_id, "FieldsList",
                                         mesh_field_threshol_id)

        # Set the mesh field as the background mesh field
        gmsh.model.mesh.field.setAsBackgroundMesh(combined_mesh_field_id)

        # Set the meshing algorithm: Quasi-structured Quad method
        gmsh.option.setNumber("Mesh.Algorithm", 11)

        # Set mesh smoothing to improve mesh quality
        gmsh.option.set_number("Mesh.Smoothing", 1000)

        # Generate the mesh in 2D
        gmsh.model.mesh.generate(dim=2)

    def write_to_msh(self, msh_path: str) -> None:
        """Writes the GMSH mesh to MSH file.

        Args:
            msh_path (str): The mesh file path in MSH format.
        """
        self._create_mesh_with_gmsh()

        gmsh.write(msh_path)
        gmsh.finalize()



"""Wind terrain scenario for air flowing over complex terrains."""

from functools import singledispatchmethod
import io
import os
import shutil
from typing import List, Optional

import numpy as np

import inductiva
from inductiva import fluids, simulation, resources, scenarios, world
from inductiva.utils import templates

SCENARIO_TEMPLATE_DIR = os.path.join(inductiva.utils.templates.TEMPLATES_PATH,
                                     "wind_terrain")
OPENFOAM_TEMPLATE_INPUT_DIR = "openfoam"
FILES_SUBDIR = "files"
COMMANDS_TEMPLATE_FILE_NAME = "commands.json.jinja"
TERRAIN_FILENAME = "terrain.stl"


class WindOverTerrain(scenarios.Scenario):
    """Wind flowing over complex terrain scenario.

    This simulation scenario models the steady-state conditions of
    wind flowing over complex terrain (e.g., mountains, valleys, etc.).

    The terrain is modeled through a 2D surface in a 3D world, defined
    through a mesh. The wind is injected through one of the side walls with
    a certain velocity vector and and at a certain region of the wall.
    Here, the wind is initialized only on a circular region of the wall
    to simulate gusts of wind.

    Schematic representation of the wind at the input wall and
    the simulation scenario both on a 2D xz-planexz (z is the vertical axis):

         Wind Injection                 2D Representation of Simulation
     ____________________                     ____________________
    |                    |                   |                    |
    |    * *             |                   |                    |
    |  *     *           |                   |->                  |
    |  *     *           |                   |->          /|      |
    |    * *             |                   |         /|/ |      |
    |____________________|                   |________/____|______|

    This scenario solves the steady-state continuity and momentum equations
    (time-independent) with the assumption of incompressible flow.
    The simulation solves the time-independent equations for several
    time steps, based on the state of the previous one. The end goal is
    to determine the steady-state of the system, i.e., where the flow
    does not change in time anymore.

    This scenario can be simulated with OpenFOAM.
    """

    valid_simulators = [fluids.simulators.OpenFOAM]

    def __init__(self,
                 terrain: world.Terrain,
                 wind_velocity: List[float],
                 wind_position: Optional[List[float]] = None,
                 atmosphere_height: float = 300):
        """Initializes the `WindTerrain` conditions.

        Args:
            wind_velocity (List): Velocity of the air flow (m/s).
            wind_position (List): Absolute Position of the wind flow (m).
                Note: The position needs to be above the terrain to occur
                any wind flow.
            terrain (inductiva.world.Terrain): Terrain object that describes
                the profile of the terrain with a mesh.
            atmosphere_height (float): Altitude (m) above the lowest point of
                terrain (m) that establishes the region of air where wind flows.
                Notice that the wind_position needs to be inside
                this atmosphere region.
        """

        self.wind_velocity = np.array(wind_velocity)
        # Compute the wind_direction from the velocity vector.
        self.wind_direction = self.wind_velocity / np.linalg.norm(
            self.wind_velocity)
        if wind_position is None:
            wind_position = [
                terrain.center["x"], terrain.center["y"],
                terrain.bounds["z"][1] + 10
            ]

        self.wind_position = wind_position
        self.terrain = terrain

        top_boundary_height = terrain.bounds["z"][0] + atmosphere_height

        if terrain.bounds["z"][1] >= top_boundary_height:
            raise ValueError(
                "The terrain height surpasses the top athmosphere boundary.")
        else:
            self.top_boundary_height = top_boundary_height

    def simulate(
        self,
        simulator: simulation.Simulator = fluids.simulators.OpenFOAM(),
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        n_cores: int = 1,
        num_iterations: int = 100,
    ) -> inductiva.tasks.Task:
        """Simulates the wind over the terrain scenario.

        Args:
            simulator: Simulator used to simulate the scenario.
                Valid simulators: OpenFOAM.
            machine_group: The MachineGroup to use for the simulation.
            num_iterations: Number of iterations to run the simulation.
            n_cores: Number of cores to use for the simulation.
            run_async: Whether to run the simulation asynchronously.
        """

        self.num_iterations = num_iterations
        self.n_cores = n_cores

        commands = self.get_commands()

        task = super().simulate(simulator,
                                machine_group=machine_group,
                                run_async=run_async,
                                n_cores=n_cores,
                                commands=commands)

        task.set_output_class(fluids.post_processing.SteadyStateOutput)

        return task

    def get_commands(self):
        """Returns the commands for the simulation."""

        commands_template_path = os.path.join(SCENARIO_TEMPLATE_DIR,
                                              OPENFOAM_TEMPLATE_INPUT_DIR,
                                              COMMANDS_TEMPLATE_FILE_NAME)

        inmemory_file = io.StringIO()
        templates.replace_params_in_template(
            template_path=commands_template_path,
            params={"n_cores": self.n_cores},
            output_file=inmemory_file,
        )
        commands = self.read_commands_from_file(inmemory_file)

        return commands

    @singledispatchmethod
    def create_input_files(self, simulator: simulation.Simulator):
        pass


@WindOverTerrain.create_input_files.register
def _(self, simulator: fluids.simulators.OpenFOAM, input_dir):  # pylint: disable=unused-argument
    """Creates OpenFOAM simulation input files."""

    # The WindTunnel with OpenFOAM requires changing multiple files
    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      OPENFOAM_TEMPLATE_INPUT_DIR, FILES_SUBDIR)

    # Copy all files from the template dir to the input directory
    shutil.copytree(template_files_dir,
                    input_dir,
                    dirs_exist_ok=True,
                    symlinks=True)

    templates.batch_replace_params_in_template(
        templates_dir=input_dir,
        template_filenames=[
            os.path.join("system", "blockMeshDict_template.openfoam.jinja"),
            os.path.join("system", "controlDict_template.openfoam.jinja"),
            os.path.join("system", "decomposeParDict_template.openfoam.jinja"),
            os.path.join("system", "snappyHexMeshDict_template.openfoam.jinja"),
            os.path.join("system", "topoSetDict_template.openfoam.jinja"),
            os.path.join("0", "include",
                         "ABLConditions_template.openfoam.jinja"),
            os.path.join("constant", "fvOptions_template.openfoam.jinja")
        ],
        params={
            "terrain_bounds": self.terrain.bounds,
            "terrain_center": self.terrain.center,
            "terrain_path": TERRAIN_FILENAME,
            "wind_flow_dir": self.wind_direction,
            "flow_position": self.wind_position,
            "top_boundary_height": self.top_boundary_height,
            "num_iterations": self.num_iterations,
            "n_cores": self.n_cores,
        },
        output_filename_paths=[
            os.path.join(input_dir, "system", "blockMeshDict"),
            os.path.join(input_dir, "system", "controlDict"),
            os.path.join(input_dir, "system", "decomposeParDict"),
            os.path.join(input_dir, "system", "snappyHexMeshDict"),
            os.path.join(input_dir, "system", "topoSetDict"),
            os.path.join(input_dir, "0", "include", "ABLConditions"),
            os.path.join(input_dir, "constant", "fvOptions")
        ],
        remove_templates=True,
    )

    # Add terrain path to its respective place
    terrain_dir = os.path.join(input_dir, "constant", "triSurface")
    os.mkdir(terrain_dir)
    terrain_file_path = os.path.join(terrain_dir, TERRAIN_FILENAME)
    self.terrain.to_text_file(terrain_file_path)



"""Describes the physical scenarios and runs its simulation via API."""

from functools import singledispatchmethod
import os
from typing import List, Literal, Optional
import shutil

from inductiva import tasks, resources
from inductiva.scenarios import Scenario
from inductiva.simulation import Simulator
from inductiva.fluids.fluid_types import FluidType
from inductiva.fluids.simulators import SPlisHSPlasH
from inductiva.fluids.simulators import DualSPHysics
from inductiva.utils.templates import replace_params_in_template
from inductiva.fluids.scenarios._post_processing import SPHSimulationOutput

TANK_DIMENSIONS = [1, 1, 1]

SPLISHSPLASH_TEMPLATE_FILENAME = "fluid_block_template.splishsplash.json.jinja"
SPLISHSPLASH_CONFIG_FILENAME = "fluid_block.json"
UNIT_BOX_MESH_FILENAME = "unit_box.obj"

DUALSPHYSICS_TEMPLATE_FILENAME = "fluid_block_template.dualsphysics.xml.jinja"
DUALSPHYSICS_CONFIG_FILENAME = "dam_break.xml"


class FluidBlock(Scenario):
    """Fluid block scenario.

    This is a simulation scenario for a fluid block moving in a cubic tank under
    the action of gravity. The tank is a cube of dimensions 1 x 1 x 1 m. The
    fluid block is also cubic, but has configurable dimensions and initial
    position and velocity. The fluid properties such as density and kinematic
    viscosity are also configurable.

    Schematic representation of the simulation scenario:
    _________________________________
    |                               |  tank
    |         ___________           |
    |         |         |           |
    |         |  fluid  |  ->       |
    |         |  block  |  initial  |
    |         |_________|  velocity |
    |                               |
    |                               |
    |_______________________________|

    The scenario can be simulated with SPlisSPlasH and DualSPHysics.
    """

    valid_simulators = [SPlisHSPlasH, DualSPHysics]

    def __init__(self,
                 density: float,
                 kinematic_viscosity: float,
                 dimensions: List[float],
                 position: Optional[List[float]] = None,
                 initial_velocity: Optional[List[float]] = None):
        """Initializes the fluid block scenario.

        Args:
            density: The density of the fluid in kg/m^3.
            kinematic_viscosity: The kinematic viscosity of the fluid, in m^2/s.
            dimensions: The fluid block dimensions (in x, y, z), in meters.
            position: The position of the fluid block in the tank (in x, y, z),
              in meters.
            initial_velocity: The initial velocity of the fluid block (in x, y,
              z), in m/s.
        """

        self.fluid = FluidType(density=density,
                               kinematic_viscosity=kinematic_viscosity)

        if len(dimensions) != 3:
            raise ValueError("`fluid_dimensions` must have 3 values.")
        self.dimensions = dimensions

        if position is None:
            self.position = [0, 0, 0]
        else:
            self.position = position

        if initial_velocity is None:
            self.initial_velocity = [0, 0, 0]
        else:
            self.initial_velocity = initial_velocity

    def simulate(
        self,
        simulator: Simulator = DualSPHysics(),
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        device: Literal["cpu", "gpu"] = "gpu",
        particle_radius: float = 0.02,
        simulation_time: float = 1,
        adaptive_time_step: bool = True,
        particle_sorting: bool = True,
        time_step: float = 0.001,
        output_time_step: float = 1 / 60,
    ) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: The simulator to use for the simulation. Supported
              simulators are: SPlisHSPlasH, DualSPHysics.
            machine_group: The machine group to use for the simulation.
            device: Device in which to run the simulation. Available options are
              "cpu" and "gpu".
            particle_radius: Radius of the fluid particles, in meters.
              Determines the resolution of the simulation. Lower values result
              in higher resolution and longer simulation times.
            simulation_time: The simulation time, in seconds.
            adaptive_time_step: Whether to use adaptive time stepping.
            particle_sorting: Whether to use particle sorting.
            time_step: Time step, in seconds.
            output_time_step: Time step between outputs, in seconds.
            run_async: Whether to run the simulation asynchronously.
        """

        # TODO: Avoid storing these as class attributes.
        self.particle_radius = particle_radius
        self.simulation_time = simulation_time
        self.adaptive_time_step = adaptive_time_step
        self.particle_sorting = particle_sorting
        self.time_step = time_step
        self.output_time_step = output_time_step

        task = super().simulate(simulator=simulator,
                                machine_group=machine_group,
                                run_async=run_async,
                                device=device)

        # TODO: Add any kind of post-processing here, e.g. convert files?
        # convert_vtk_data_dir_to_netcdf(
        #     data_dir=os.path.join(output_path, "vtk"),
        #     output_time_step=SPLISHSPLASH_OUTPUT_TIM_STEP,
        #     netcdf_data_dir=os.path.join(output_path, "netcdf"))
        task.set_output_class(SPHSimulationOutput)

        return task

    @singledispatchmethod
    def get_config_filename(self, simulator: Simulator):
        pass

    @singledispatchmethod
    def create_input_files(self, simulator: Simulator):
        pass


@FluidBlock.get_config_filename.register
def _(cls, simulator: SPlisHSPlasH):  # pylint: disable=unused-argument
    """Returns the configuration filename for SPlisHSPlasH."""
    return SPLISHSPLASH_CONFIG_FILENAME


@FluidBlock.create_input_files.register
def _(self, simulator: SPlisHSPlasH, input_dir):  # pylint: disable=unused-argument
    """Creates SPlisHSPlasH simulation input files."""

    # Copy the unit box mesh file to the input directory.
    unit_box_file_path = os.path.join(os.path.dirname(__file__),
                                      UNIT_BOX_MESH_FILENAME)
    shutil.copy(unit_box_file_path, input_dir)

    # Generate the simulation configuration file.
    fluid_margin = 2 * self.particle_radius

    replace_params_in_template(
        template_path=os.path.join(os.path.dirname(__file__),
                                   SPLISHSPLASH_TEMPLATE_FILENAME),
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_radius": self.particle_radius,
            "data_export_rate": 1 / self.output_time_step,
            "tank_filename": UNIT_BOX_MESH_FILENAME,
            "tank_dimensions": TANK_DIMENSIONS,
            "fluid_filename": UNIT_BOX_MESH_FILENAME,
            "fluid": self.fluid,
            "fluid_position": [
                position + fluid_margin for position in self.position
            ],
            "fluid_dimensions": [
                dimension - 2 * fluid_margin for dimension in self.dimensions
            ],
            "fluid_velocity": self.initial_velocity,
            "particle_sorting": self.particle_sorting,
            "adaptive_time_step": self.adaptive_time_step,
        },
        output_file=os.path.join(input_dir, SPLISHSPLASH_CONFIG_FILENAME),
    )


@FluidBlock.get_config_filename.register
def _(cls, simulator: DualSPHysics):  # pylint: disable=unused-argument
    """Returns the configuration filename for DualSPHysics."""
    return DUALSPHYSICS_CONFIG_FILENAME


@FluidBlock.create_input_files.register
def _(self, simulator: DualSPHysics, input_dir):  # pylint: disable=unused-argument
    """Creates DualSPHysics simulation input files."""

    replace_params_in_template(
        template_path=os.path.join(os.path.dirname(__file__),
                                   DUALSPHYSICS_TEMPLATE_FILENAME),
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_distance": 2 * self.particle_radius,
            "output_time_step": self.output_time_step,
            "tank_dimensions": TANK_DIMENSIONS,
            "fluid_dimensions": self.dimensions,
            "fluid_position": self.position,
            "fluid": self.fluid,
            "adaptive_time_step": self.adaptive_time_step,
        },
        output_file=os.path.join(input_dir, DUALSPHYSICS_CONFIG_FILENAME),
    )



"""Visualization processing of WindTunnel scenario.
Currently, we only support the OpenFOAM simulator.
"""
import os
import csv

from inductiva.fluids import post_processing
from inductiva import types


class WindTunnelOutput(post_processing.SteadyStateOutput):
    """Post-Process WindTunnel simulation outputs.
    This class is based on a more general class for post-processing the
    outputs of steady-state simulations. It inherits the following tools:
        - Pressure over object;
        - Cutting plane;
        - Stream lines;

    Additionally, it provides the following methods:
        - Force coefficients (added here only for this scenario).

    Current Support:
        OpenFOAM.
    """

    def get_force_coefficients(self, save_path: types.Path = None):
        """Get the force coefficients of the object in the WindTunnel.
        
        The force coefficients are provided in a .dat file during the
        simulation run-time. This file contains 8 lines that are provide
        the general input information. In this function, we read the file,
        ignore the first 8 lines and read the force coefficients for the 
        simulation_time chosen.

        Args:
            save_path: Path to save the force coefficients in a .csv file.
        """

        if not self.full_output:
            force_coefficients_path = os.path.join(self.sim_output_path,
                                                   "force_coefficients.csv")
            with open(force_coefficients_path, "r",
                      encoding="utf-8") as csv_file:
                csv_reader = csv.reader(csv_file)
                force_coefficients = list(csv_reader)
        else:
            num_header_lines = 8
            force_coefficients_path = os.path.join(self.sim_output_path,
                                                   "postProcessing",
                                                   "forceCoeffs1", "0",
                                                   "forceCoeffs.dat")
            force_coefficients = []
            last_iteration = int(self.get_last_iteration())

            with open(force_coefficients_path, "r",
                      encoding="utf-8") as forces_file:
                for index, line in enumerate(forces_file.readlines()):
                    # Pick the line 8 of the file:
                    # [#, Time, Cm, Cd, Cl, Cl(f), Cl(r)] and remove # column
                    if index == num_header_lines:
                        force_coefficients.append(line.split()[1:])
                    # Add the force coefficients for the simulation time chosen
                    elif index == num_header_lines + last_iteration + 1:
                        force_coefficients.append(line.split())

            if save_path:
                with open(save_path, "w", encoding="utf-8") as csv_file:
                    csv_writer = csv.writer(csv_file)
                    csv_writer.writerows(force_coefficients)

        return force_coefficients



"""Functions to manage or retrieve user resources."""
from absl import logging
import inductiva
from inductiva.client.apis.tags import instance_api
from inductiva import resources


def list_machine_groups():
    """Lists all active machine group names.

    Returns:
        List of machine group names.
    """
    try:
        api = instance_api.InstanceApi(inductiva.api.get_client())
        response = api.list_active_user_instance_groups()
        if len(response.body) == 0:
            logging.info("No active machine groups found.")
            return response.body

        machine_group_names = []
        for machine_group in response.body:
            machine_group_names.append(machine_group["name"])
            logging.info("Name: %s; Number of machines: %s; Created at: %s",
                         machine_group["name"], machine_group["num_instances"],
                         machine_group["create_time"])
        return machine_group_names

    except inductiva.client.ApiException as api_exception:
        raise api_exception


def get_machine_group(name: str):
    """Returns a 'MachineGroup' object.

    Given the name of the machine group, returns a 'MachineGroup' object
    with the attributes of that machine group.

    Args:
        name: Name of the machine group."""
    try:
        api = instance_api.InstanceApi(inductiva.api.get_client())
        response = api.get_instance_group({"name": name})

        mg = resources.MachineGroup(machine_type=response.body["machine_type"],
                                    num_machines=response.body["num_instances"],
                                    spot=response.body["spot"],
                                    disk_size_gb=response.body["disk_size_gb"],
                                    zone=response.body["zone"])
        mg.name = response.body["name"]
        return mg

    except inductiva.client.ApiException as api_exception:
        raise api_exception



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class GCloudHostInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Info about Google Cloud machine hosting the executer.
    """

    class MetaOapg:
        required = {
            "vm_id",
            "vm_type",
            "preemptible",
            "vm_metadata",
            "host_type",
            "vm_name",
        }

        class properties:

            class host_type(schemas.EnumBase, schemas.StrSchema):

                class MetaOapg:
                    enum_value_to_name = {
                        "gcloud": "GCLOUD",
                    }

                @schemas.classproperty
                def GCLOUD(cls):
                    return cls("gcloud")

            vm_type = schemas.StrSchema
            vm_name = schemas.StrSchema
            vm_id = schemas.StrSchema
            preemptible = schemas.BoolSchema
            vm_metadata = schemas.DictSchema
            __annotations__ = {
                "host_type": host_type,
                "vm_type": vm_type,
                "vm_name": vm_name,
                "vm_id": vm_id,
                "preemptible": preemptible,
                "vm_metadata": vm_metadata,
            }

    vm_id: MetaOapg.properties.vm_id
    vm_type: MetaOapg.properties.vm_type
    preemptible: MetaOapg.properties.preemptible
    vm_metadata: MetaOapg.properties.vm_metadata
    host_type: MetaOapg.properties.host_type
    vm_name: MetaOapg.properties.vm_name

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["host_type"]
    ) -> MetaOapg.properties.host_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["vm_type"]
    ) -> MetaOapg.properties.vm_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["vm_name"]
    ) -> MetaOapg.properties.vm_name:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["vm_id"]
    ) -> MetaOapg.properties.vm_id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["preemptible"]
    ) -> MetaOapg.properties.preemptible:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["vm_metadata"]
    ) -> MetaOapg.properties.vm_metadata:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "host_type",
        "vm_type",
        "vm_name",
        "vm_id",
        "preemptible",
        "vm_metadata",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["host_type"]
    ) -> MetaOapg.properties.host_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["vm_type"]
    ) -> MetaOapg.properties.vm_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["vm_name"]
    ) -> MetaOapg.properties.vm_name:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["vm_id"]
    ) -> MetaOapg.properties.vm_id:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["preemptible"]
    ) -> MetaOapg.properties.preemptible:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["vm_metadata"]
    ) -> MetaOapg.properties.vm_metadata:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "host_type",
        "vm_type",
        "vm_name",
        "vm_id",
        "preemptible",
        "vm_metadata",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        vm_id: typing.Union[
            MetaOapg.properties.vm_id,
            str,
        ],
        vm_type: typing.Union[
            MetaOapg.properties.vm_type,
            str,
        ],
        preemptible: typing.Union[
            MetaOapg.properties.preemptible,
            bool,
        ],
        vm_metadata: typing.Union[
            MetaOapg.properties.vm_metadata,
            dict,
            frozendict.frozendict,
        ],
        host_type: typing.Union[
            MetaOapg.properties.host_type,
            str,
        ],
        vm_name: typing.Union[
            MetaOapg.properties.vm_name,
            str,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'GCloudHostInfo':
        return super().__new__(
            cls,
            *_args,
            vm_id=vm_id,
            vm_type=vm_type,
            preemptible=preemptible,
            vm_metadata=vm_metadata,
            host_type=host_type,
            vm_name=vm_name,
            _configuration=_configuration,
            **kwargs,
        )



"""Fluid tank scenario."""
from dataclasses import dataclass, field
from enum import Enum
from functools import singledispatchmethod
import os
from typing import List, Literal, Optional

from inductiva import tasks, resources
from inductiva.scenarios import Scenario
from inductiva.simulation import Simulator
from inductiva.fluids.shapes import BaseShape
from inductiva.fluids.shapes import Rectangle
from inductiva.fluids.shapes import Circle
from inductiva.fluids.shapes import Cube
from inductiva.fluids.shapes import Cylinder
from inductiva.fluids.fluid_types import FluidType
from inductiva.fluids.fluid_types import WATER
from inductiva.fluids.simulators import SPlisHSPlasH
from inductiva.utils.templates import replace_params_in_template
from inductiva.fluids.scenarios.fluid_tank.output import FluidTankOutput

from . import mesh_file_utils

SPLISHSPLASH_TEMPLATE_FILENAME = "fluid_tank_template.splishsplash.json.jinja"
SPLISHSPLASH_CONFIG_FILENAME = "fluid_tank.json"
TANK_MESH_FILENAME = "tank.obj"
FLUID_MESH_FILENAME = "fluid.obj"


@dataclass
class ParticleRadius(Enum):
    """Sets particle radius according to resolution."""
    HIGH = 0.005
    MEDIUM = 0.01
    LOW = 0.02


@dataclass
class TimeStep(Enum):
    """Sets time step according to resolution."""
    HIGH = 0.0025
    MEDIUM = 0.005
    LOW = 0.01


# Tank inlets.
@dataclass
class BaseTankInlet:
    """Base tank inlet."""
    fluid_velocity: float = 1
    position: List[float] = field(default_factory=lambda: [0, 0])


@dataclass
class RectangularTankInlet(BaseTankInlet, Rectangle):
    """Rectangular tank inlet."""
    pass


@dataclass
class CircularTankInlet(BaseTankInlet, Circle):
    """Circular tank inlet."""
    pass


# Tank outlets.
class BaseTankOutlet:
    """Base tank outlet."""
    pass


class CubicTankOutlet(BaseTankOutlet):
    """Cubic tank outlet."""

    def __init__(
        self,
        dimensions: List[float],
        top_base_position: Optional[List[float]] = None,
    ):
        """Initializes a cubic tank outlet.

        Args:
            dimensions: Dimensions of the outlet.
            top_base_position: Position of the top base of the outlet.
        """
        top_base_position = top_base_position or [0, 0]

        self.shape = Cube(
            dimensions=dimensions,
            position=[*top_base_position, -dimensions[2]],
        )


class CylindricalTankOutlet(BaseTankOutlet):
    """Cylindrical tank outlet."""

    def __init__(
        self,
        radius: float,
        height: float,
        top_base_position: Optional[List[float]] = None,
    ):
        """Initializes a cylindrical tank outlet.

        Args:
            radius: Radius of the outlet.
            height: Height of the outlet.
            top_base_position: Position of the top base of the outlet.
        """
        top_base_position = top_base_position or [0, 0]

        self.shape = Cylinder(
            radius=radius,
            height=height,
            position=[*top_base_position, -height],
        )


class FluidTank(Scenario):
    """Fluid tank scenario.

    This is a simulation scenario for a fluid tank. The tank has a 3D shape that
    may be cubic or cylindrical. Fluid is injected in the tank via an inlet
    located at the top of the tank, and flows out of the tank via an outlet
    located at the bottom of the tank. The motion of the fluid is controled
    by gravity. The fluid properties such as density and kinematic viscosity
    are configurable. The initial fluid level in the tank is also configurable,
    as well as the inlet and outlet positions and dimensions.

    The main axis of the tank is the z axis. The inlet and outlet are located at
    the top and bottom of the tank, respectively, along the z axis. Fluid is
    injected from the inlet in the negative z direction with a given velocity.

    Schematic representation of the simulation scenario: e.g. x/y points right,
    z points up.

       inlet
    _____________________
    |    |              |
    |    v              |
    |                   |
    |                   |
    |___________________|  fluid level
    |                   |
    |                   |
    |                   |
    |                   |
    |___________________|
                    |
                    v  outlet

    The scenario can be simulated with SPlisHSPlasH.
    """

    valid_simulators = [SPlisHSPlasH]

    def __init__(
        self,
        shape: BaseShape = Cylinder(radius=0.5, height=1),
        fluid: FluidType = WATER,
        fluid_level: float = 0.5,
        inlet: Optional[BaseTankInlet] = CircularTankInlet(radius=0.1),
        outlet: Optional[BaseTankOutlet] = CylindricalTankOutlet(radius=0.1,
                                                                 height=0.1),
    ):
        """Initializes a fluid tank.

        Args:
            shape: The shape of the tank.
            fluid: The fluid type.
            fluid_level: The fluid level initially in the tank.
            inlet: The inlet of the tank.
            outlet: The outlet of the tank.
        """
        self.shape = shape
        self.fluid = fluid
        self.fluid_level = fluid_level
        self.inlet = inlet
        self.outlet = outlet

    def simulate(
        self,
        simulator: Simulator = SPlisHSPlasH(),
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        device: Literal["cpu", "gpu"] = "cpu",
        simulation_time: float = 5,
        resolution: Literal["low", "medium", "high"] = "low",
        output_time_step: float = 0.1,
        particle_sorting: bool = False,
    ) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: Simulator to use. Supported simulators are: SPlisHSPlasH.
            machine_group: The machine group to use for the simulation.
            simulation_time: Total simulation time, in seconds.
            output_time_step: Time step for the output, in seconds.
            resolution: Resolution of the simulation. Controls the particle
              radius and time step. Accepted values are: "low", "medium",
              "high".
            particle_sorting: Whether to use particle sorting.
            run_async: Whether to run the simulation asynchronously.
        """

        self.simulation_time = simulation_time
        self.particle_radius = ParticleRadius[resolution.upper()].value
        self.time_step = TimeStep[resolution.upper()].value
        self.output_time_step = output_time_step
        self.particle_sorting = particle_sorting

        task = super().simulate(
            simulator,
            machine_group=machine_group,
            run_async=run_async,
            device=device,
            sim_config_filename=self.get_config_filename(simulator),
        )

        task.set_output_class(FluidTankOutput)

        return task

    def get_bounding_box(self):
        """Gets the bounding box of the tank.

        Returns:
            Tuple of two lists representing the minimum and maximum coordinates
            of the bounding box of the tank, respectively.
        """

        bounding_box_min, bounding_box_max = self.shape.get_bounding_box()

        # Extend the bounding box to include the outlet.
        if self.outlet is not None:
            outlet_bounding_box_min, _ = self.outlet.shape.get_bounding_box()
            bounding_box_min[2] = outlet_bounding_box_min[2]

        return bounding_box_min, bounding_box_max

    def create_mesh_files(self, output_dir: str):
        """Creates mesh files for the tank and fluid in the given directory."""
        mesh_file_utils.create_tank_mesh_file(
            shape=self.shape,
            outlet=self.outlet,
            path=os.path.join(output_dir, TANK_MESH_FILENAME),
        )

        mesh_file_utils.create_tank_fluid_mesh_file(
            shape=self.shape,
            fluid_level=self.fluid_level,
            margin=2 * self.particle_radius,
            path=os.path.join(output_dir, FLUID_MESH_FILENAME),
        )

    @singledispatchmethod
    def get_config_filename(self, simulator: Simulator):
        pass

    @singledispatchmethod
    def create_input_files(self, simulator: Simulator):
        pass


@FluidTank.get_config_filename.register
def _(cls, simulator: SPlisHSPlasH) -> str:  # pylint: disable=unused-argument
    """Returns the config filename for SPlisHSPlasH."""
    return SPLISHSPLASH_CONFIG_FILENAME


@FluidTank.create_input_files.register
def _(self, simulator: SPlisHSPlasH, input_dir):  # pylint: disable=unused-argument
    """Creates SPlisHSPlasH simulation input files."""

    self.create_mesh_files(input_dir)

    bounding_box_min, bounding_box_max = self.get_bounding_box()
    inlet_position = [
        self.inlet.position[0],
        self.inlet.position[1],
        bounding_box_max[2],
    ]

    replace_params_in_template(
        template_path=os.path.join(os.path.dirname(__file__),
                                   SPLISHSPLASH_TEMPLATE_FILENAME),
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_radius": self.particle_radius,
            "data_export_rate": 1 / self.output_time_step,
            "z_sort": self.particle_sorting,
            "tank_filename": TANK_MESH_FILENAME,
            "fluid_filename": FLUID_MESH_FILENAME,
            "fluid": self.fluid,
            "inlet_position": inlet_position,
            "inlet_width": int(self.inlet.radius / self.particle_radius),
            "inlet_fluid_velocity": self.inlet.fluid_velocity,
            "bounding_box_min": bounding_box_min,
            "bounding_box_max": bounding_box_max,
        },
        output_file=os.path.join(input_dir, SPLISHSPLASH_CONFIG_FILENAME),
    )



"""Utility functions to manipulate grid data."""
import typing

import scipy
import numpy as np


def get_meshgrid(x_range: typing.Sequence[float],
                 y_range: typing.Sequence[float], x_num: int, y_num: int):

    x_grid = np.linspace(*x_range, num=x_num)
    y_grid = np.linspace(*y_range, num=y_num)

    x_grid, y_grid = np.meshgrid(x_grid, y_grid, indexing="ij")

    return x_grid, y_grid


def reshape_map(x_num: int, y_num: int, map_level: np.ndarray):
    """Reshape a map into a different resolution.
    
    Args:
        x_num: New number of grid points in x-direction.
        y_num: New number of grid points in y-direction.
        map_level: Values on the z-axis of the previous grid.
    """

    x_grid_prev, y_grid_prev = np.meshgrid(np.arange(map_level.shape[0]),
                                           np.arange(map_level.shape[1]),
                                           indexing="ij")

    x_grid_new, y_grid_new = np.meshgrid(np.arange(x_num),
                                         np.arange(y_num),
                                         indexing="ij")

    return scipy.interpolate.griddata(
        (x_grid_prev.flatten(), y_grid_prev.flatten()),
        map_level.flatten(),
        (x_grid_new, y_grid_new),
        method="linear",
    )



"""Base class for low-level simulators."""
from typing import Optional
from abc import ABC, abstractmethod

from inductiva import types, tasks, resources
from inductiva.utils import files


class Simulator(ABC):
    """Base simulator class."""

    @property
    @abstractmethod
    def api_method_name(self) -> str:
        pass

    def _setup_input_dir(self, input_dir: types.Path):
        """Setup the simulator input directory."""
        input_dir = files.resolve_path(input_dir)
        if not input_dir.is_dir():
            raise ValueError(
                f"The provided path (\"{input_dir}\") is not a directory.")
        return input_dir

    def run(
        self,
        input_dir: types.Path,
        *_args,
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        **kwargs,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory containing the input files.
            _args: Unused in this method, but defined to allow for more
                non-default arguments in method override in subclasses.
            **kwargs: Additional keyword arguments to be passed to the
                simulation API method.
        """
        input_dir = self._setup_input_dir(input_dir)

        return tasks.run_simulation(
            self.api_method_name,
            input_dir,
            run_async=run_async,
            machine_group=machine_group,
            **kwargs,
        )



"""Functions for running simulations via Inductiva Web API."""
import pathlib
from typing import Any, Optional

from inductiva import tasks, resources
from inductiva.api import methods


def run_simulation(
    api_method_name: str,
    input_dir: pathlib.Path,
    machine_group: Optional[resources.MachineGroup] = None,
    run_async: bool = False,
    **kwargs: Any,
) -> tasks.Task:
    """Run a simulation via Inductiva Web API."""

    params = {
        "sim_dir": input_dir,
        **kwargs,
    }
    type_annotations = {
        "sim_dir": pathlib.Path,
    }

    resource_pool_id = None
    if machine_group is not None:
        resource_pool_id = machine_group.id

    task_id = methods.invoke_async_api(api_method_name,
                                       params,
                                       type_annotations,
                                       resource_pool_id=resource_pool_id)
    task = tasks.Task(task_id)
    if not isinstance(task_id, str):
        raise RuntimeError(
            f"Expected result to be a string with task_id, got {type(task_id)}")

    # Blocking call for sync execution
    if not run_async:
        with task:
            task.wait()

    return task



# pylint: disable=missing-module-docstring
import re


def split_camel_case(s):
    """Split a camel case string into a list of strings."""
    return re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1",
                                                  s)).split()


def format_bytes(n_bytes: int) -> str:
    """Convert bytes to human readable string."""
    res = float(n_bytes)

    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if res < 1024:
            if unit == "B":
                return f"{res:.0f} {unit}"
            else:
                return f"{res:.2f} {unit}"
        res /= 1024

    return f"{format_bytes:.2f} PiB"



"Postprocessing steps for the MDWaterBox scenario."
import os
import nglview as nv
from pathlib import Path
from ..utils import unwrap_trajectory


class MDWaterBoxOutput:
    """Post process the simulation output of a MDWaterBox scenario."""

    def __init__(self, sim_output_path: Path = None):
        """Initializes a `MDWaterBoxOutput` object.

        Given a simulation output directory that contains the standard files
        generated by a MDWaterBox simulation run, this class provides methods to
        visualize the simulation outputs in a notebook interactively.

        Args:
            sim_output_path: Path to the simulation output directory."""

        self.sim_output_dir = sim_output_path

    def render_interactive(self, use_compressed_trajectory: bool = False):
        """Render the simulation outputs in an interactive visualization.
        Args: 
            use_compressed_trajectory: Whether to use the compressed trajectory
            or the full precision trajectory."""
        topology = os.path.join(self.sim_output_dir, "eql.tpr")
        if use_compressed_trajectory:
            trajectory = os.path.join(self.sim_output_dir, "trajectory.xtc")
        else:
            trajectory = os.path.join(self.sim_output_dir, "eql.trr")
        universe = unwrap_trajectory(topology, trajectory)
        view = nv.show_mdanalysis(universe)
        view.add_ball_and_stick("all")
        view.center()
        print("System Information:")
        print(f"Number of atoms in the system: {len(universe.atoms)}")
        print(f"Number of trajectory frames: {len(universe.trajectory)}")
        return view



"""Molecular Dynamics simulation for water box scenario."""
from functools import singledispatchmethod
from typing import Optional, Literal
import os
import shutil
import io

from inductiva import tasks, resources
from inductiva.molecules.simulators import GROMACS
from inductiva.simulation import Simulator
from inductiva.utils import templates
from inductiva.scenarios import Scenario
from .post_processing import MDWaterBoxOutput

SCENARIO_TEMPLATE_DIR = os.path.join(templates.TEMPLATES_PATH, "md_water_box")
GROMACS_TEMPLATE_INPUT_DIR = "gromacs"
COMMANDS_TEMPLATE_FILE_NAME = "commands.json.jinja"


class MDWaterBox(Scenario):
    """Molecular dynamics water box scenario."""

    valid_simulators = [GROMACS]

    def __init__(
        self,
        temperature: float = 300,
        box_size: float = 2.3,
    ):
        """The scenario involves simulating a box with water molecules.
        The simulation consists of two main steps: energy minimization
        and molecular dynamics simulation.
        By default, the initial positions of the water molecules are
        arranged uniformally, so to randomize them we perform a
        decorrelation step.
        Args:
            temperature: The temperature of the simulation in Kelvin.
            box_size: The size of the box in nm.
        """

        self.temperature = temperature
        if box_size < 2.3:
            raise ValueError("The box size must be greater than 2.3 nm.")
        self.box_size = box_size

    def simulate(
            self,
            simulator: Simulator = GROMACS(),
            machine_group: Optional[resources.MachineGroup] = None,
            run_async: bool = False,
            simulation_time_ns: float = 10,  # ns
            integrator: Literal["md", "sd", "bd"] = "md",
            n_steps_min: int = 5000) -> tasks.Task:
        """Simulate the water box scenario using molecular dynamics.

        Args:
            machine_group: The MachineGroup to use for the simulation.
            simulation_time_ns: The simulation time in ns.
            integrator: The integrator to use for the simulation. Options:
                - "md" (Molecular Dynamics): Accurate leap-frog algorithm for
                integrating Newton's equations of motion.
                - "sd" (Steepest Descent): Stochastic dynamics integrator with
                leap-frog scheme.
                - "bd" (Brownian Dynamics): Euler integrator for Brownian or
                position Langevin dynamics.

            For more details on the integrators, refer to the GROMACS
            documentation at
            https://manual.gromacs.org/current/user-guide/mdp-options.html.

            machine_group: The machine group to use for the simulation.
            n_steps_min: Number of steps for energy minimization.
            run_async: Whether to run the simulation asynchronously.
        """
        self.nsteps = int(
            simulation_time_ns * 1e6 / 2
        )  # convert to fs and divide by the time step of the simulation (2 fs)
        self.integrator = integrator
        self.n_steps_min = n_steps_min

        commands = self.get_commands()
        task = super().simulate(simulator,
                                machine_group=machine_group,
                                commands=commands,
                                run_async=run_async)

        task.set_output_class(MDWaterBoxOutput)

        return task

    def get_commands(self):
        """Returns the commands for the simulation."""

        commands_template_path = os.path.join(SCENARIO_TEMPLATE_DIR,
                                              GROMACS_TEMPLATE_INPUT_DIR,
                                              COMMANDS_TEMPLATE_FILE_NAME)

        inmemory_file = io.StringIO()
        templates.replace_params_in_template(
            template_path=commands_template_path,
            params={"box_size": self.box_size},
            output_file=inmemory_file,
        )
        commands = self.read_commands_from_file(inmemory_file)

        return commands

    @singledispatchmethod
    def create_input_files(self, simulator: Simulator):
        pass


@MDWaterBox.create_input_files.register
def _(self, simulator: GROMACS, input_dir):  # pylint: disable=unused-argument
    """Creates GROMACS simulation input files."""

    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      GROMACS_TEMPLATE_INPUT_DIR)

    shutil.copytree(template_files_dir, input_dir, dirs_exist_ok=True)

    templates.batch_replace_params_in_template(
        templates_dir=input_dir,
        template_filenames=[
            "simulation.mdp.jinja",
            "energy_minimization.mdp.jinja",
        ],
        params={
            "integrator": self.integrator,
            "nsteps": self.nsteps,
            "ref_temp": self.temperature,
            "nsteps_minin": self.n_steps_min,
        },
        output_filename_paths=[
            os.path.join(input_dir, "simulation.mdp"),
            os.path.join(input_dir, "energy_minimization.mdp"),
        ],
        remove_templates=True,
    )



"""Heat sink scenario."""
from functools import singledispatchmethod
import os
import shutil
from typing import Optional

from inductiva.fluids.scenarios.heat_sink.output import HeatSinkOutput

from inductiva import tasks, resources
from inductiva.fluids.simulators import OpenFOAM
from inductiva.simulation import Simulator
from inductiva.scenarios import Scenario
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       replace_params_in_template)

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "heat_sink")
OPENFOAM_TEMPLATE_SUBDIR = "openfoam"
FILES_SUBDIR = "files"
OPENFOAM_TEMPLATE_PARAMS_FILE_NAME = "parameters.jinja"
OPENFOAM_PARAMS_FILE_NAME = "parameters"
COMMANDS_FILE_NAME = "commands.json"


class HeatSink(Scenario):
    """Heat sink scenario.

    This is a simulation scenario for a heat sink. A heat source is placed
    in a box where there is an air flow. A heat sink, placed on top of the
    source, is used to dissipate the heat via convection with the air flow.

    The heat source is modeled as a heater with a given power. The heat sink
    is a block of aluminum with thin fins on top. The air flow is introduced
    in the simulation via an inlet, where the air is at a fixed temperature.

    The scenario is simulated in a 3D box with dimensions 8 x 16 x 52 cm in x, y
    and z directions, respectively. The heat sink is a 4 x 3 x 6 cm block
    centered in (x, z) in the simulation box. The sink has 1 mm wide fins
    elongated along the z direction and separated by 2 mm centered in the x
    direction. The heat source is a 1 x 1 x 1 cm cube that sits under the heat
    sink. The air flow is injected in the simulation from the lower z boundary,
    and flows in the positive z direction.

    Schematic representations of the simulation scenario:

    - as seen from the side (zy plane): z points right, y points up, x points
      into the screen.
      _________________________________
      |                               |
      |                               |
      |                               |
      |  air flow ->                  |
      |           _________           |
      |           |       | heat      |
      |           |       | sink      |
      |           |       |           |
      |___________|_______|___________|
                    |___|
                 heat source

    - as seen from the the inlet of the air flow (xy plane): x points right, y
      points up, z points out of the screen.

      _________________________________
      |                               |
      |                               |
      |                               |
      |                               |
      |                               |
      |      fins | | | | | heat      |
      |           | | | | | sink      |
      |           |_|_|_|_|           |
      |           |       |           |
      |___________|_______|___________|
                    |___|
                 heat source

    The scenario can be simulated with OpenFOAM.
    """

    valid_simulators = [OpenFOAM]

    def __init__(
        self,
        air_velocity=10,
        air_temperature=290,
        heater_power=200,
    ):
        """Initializes the heat sink scenario.

        Args:
            air_velocity: The velocity of the air flow, in m/s.
            air_temperature: The temperature of the air flow, in Kelvin. Also
              sets the initial temperature of the heater and the heat sink.
            heater_power: The power of the heater, in Watts.
        """
        self.air_velocity = air_velocity
        self.air_temperature = air_temperature
        self.heater_power = heater_power

    def simulate(
        self,
        simulator: Simulator = OpenFOAM(),
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        simulation_time=300,
        output_time_step=10,
    ) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: The simulator to use for the simulation.
            simulation_time: The simulation time, in seconds.
            output_time_step: The time step to save the simulation results, in
              seconds.
            machine_group: The machine group to use for the simulation.
            run_async: Whether to run the simulation asynchronously.
        """
        self.simulation_time = simulation_time
        self.output_time_step = output_time_step

        commands = self.get_commands()

        task = super().simulate(simulator,
                                machine_group=machine_group,
                                run_async=run_async,
                                commands=commands)

        task.set_output_class(HeatSinkOutput)

        return task

    def get_commands(self):
        """Returns the OpenFOAM commands for the simulation.
        """
        commands_file_path = os.path.join(SCENARIO_TEMPLATE_DIR,
                                          OPENFOAM_TEMPLATE_SUBDIR,
                                          COMMANDS_FILE_NAME)

        commands = self.read_commands_from_file(commands_file_path)

        return commands

    @singledispatchmethod
    def create_input_files(self, simulator: Simulator):
        pass


@HeatSink.create_input_files.register
def _(self, simulator: OpenFOAM, input_dir):  # pylint: disable=unused-argument
    """Creates OpenFOAM simulation input files."""

    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      OPENFOAM_TEMPLATE_SUBDIR, FILES_SUBDIR)

    shutil.copytree(template_files_dir,
                    input_dir,
                    dirs_exist_ok=True,
                    symlinks=True)

    template_file_path, params_file_path = (
        os.path.join(input_dir, file_name) for file_name in
        [OPENFOAM_TEMPLATE_PARAMS_FILE_NAME, OPENFOAM_PARAMS_FILE_NAME])

    replace_params_in_template(
        template_path=template_file_path,
        params={
            "simulation_time": self.simulation_time,
            "output_time_step": self.output_time_step,
            "air_velocity": self.air_velocity,
            "air_temperature": self.air_temperature,
            "heater_power": self.heater_power
        },
        output_file=params_file_path,
        remove_template=True,
    )



"""Tools to analyze molecular dynamics simulations."""
import MDAnalysis as mda
from MDAnalysis import transformations
from MDAnalysis.analysis import align


def unwrap_trajectory(topology, trajectory):
    """Unwrap visualization of the trajectory to deal with 
    Periodic Boundary Conditions.
    Args:
        topology: Path to the topology file.
        trajectory: Path to the trajectory file."""
    universe = mda.Universe(topology, trajectory, guess_bonds=True)
    atoms = universe.atoms
    transformation = transformations.unwrap(atoms)
    universe.trajectory.add_transformations(transformation)
    return universe


def align_trajectory_to_average(universe, trajectory_output_path):
    """Align the trajectory to the average structure. 
    Args:
        universe: The universe MDAnalysis object.
        trajectory_output_path: Path to the aligned trajectory file."""
    average = align.AverageStructure(universe,
                                     universe,
                                     select="protein and name CA",
                                     ref_frame=0).run()
    average_trajectory = average.results.universe

    align.AlignTraj(universe,
                    average_trajectory,
                    select="protein and name CA",
                    filename=trajectory_output_path,
                    in_memory=False).run()



#pylint: disable=missing-module-docstring
from .fluid_tank import (
    FluidTank,
    CubicTankOutlet,
    CylindricalTankOutlet,
    RectangularTankInlet,
    CircularTankInlet,
)
from .output import FluidTankOutput



"""Functions to generate elevation data for a general grid."""

import math
import typing

import numpy as np

import inductiva
from inductiva.generative import procedural


def generate_random_map_level(
    x_num: int,
    y_num: int,
    corner_values: typing.Sequence[float],
    initial_roughness: float = 1,
    roughness_factor: float = 0.5,
    random_seed: int = None,
):
    """Generate a set of elevation values for xy-grid.
    
    This function generalises the diamond-square algorithm for grids that are
    not square and multiple of 2^n+1.

    A grid of random elevations of shape `(x_num, y_num)` is generated as
    follows:
    1. A square grid of random elevations is generated using the Diamond-Square
    algorithm. The size of the resultnig grid is `n`, where `n` is
    the smallest integer that satisfies `2^n + 1 >= max(x_num, y_num)`.
    2. The elevations of the square grid are interpolated to a grid of shape
    `(x_num, y_num)`.

    Args:
        x_num: The number of points in the x direction.
        y_num: The number of points in the y direction.
        corner_values: Sequence of 4 values establishing the elevation of the 
            grid corners. The order refers to top-left, top-right,
            bottom-left, and bottom-right.
        initial_roughness: Initial roughness value, in meters. Controls the
            initial range of randomness of the Diamond-Square algorithm.
        roughness_factor: Roughness factor. Must be between 0 and 1.
            Controls the rate at which the range of randomness of the
            Diamond-Square algorithm decreases over iterations.
        random_seed: Random seed to use for the Diamond-Square algorithm.
    """

    # Determine the minimum n such that 2^n + 1 >= max(x_num, y_num).
    n_power = int(math.log2(max(x_num, y_num) - 1)) + 1

    size_square = 2**n_power + 1

    # Create elevation for a square grid with side resolution=size_square.
    map_level = procedural.diamond_square.create_random_array(
        size=size_square,
        corner_values=corner_values,
        initial_roughness=initial_roughness,
        roughness_factor=roughness_factor,
        random_seed=random_seed)

    return inductiva.utils.grids.reshape_map(x_num=x_num,
                                             y_num=y_num,
                                             map_level=map_level)


def adjust_map_level(map_level: np.ndarray, percentile_translate_map: float):
    """Adjust a map level based on percentile.
    
    Args:
        map_level: The map level to adjust.
        percentile_translate_map: The percentile of the map level to adjust.
        """
    # Adjust a map_level to ensure that a given percentage of it
    # is above or below the xy-plane.
    percentile_translate = np.percentile(map_level,
                                         abs(percentile_translate_map))
    map_level += np.sign(percentile_translate_map) * percentile_translate

    return map_level



"""Utils to create the boundary conditions."""

from abc import ABC, abstractmethod
from typing import Optional, Literal


class BoundaryCondition(ABC):
    """Abstract base class for boundary condition.

    Attributes:
        boundary_name (float): The boundary name of the plate where the boundary
          condition will be apply. 
          The available options are: left, top, right and bottom.
    """

    def __init__(
            self, boundary_name: Literal["left", "top", "right",
                                         "bottom"]) -> None:
        """Initializes a BoundaryCondition object."""
        self.boundary_name = boundary_name

    @abstractmethod
    def to_dict(self):
        """Abstract method to convert the bc properties to a dictionary."""
        pass


class DirichletBC(BoundaryCondition):
    """Dirichlet boundary condition.

    A Dirichlet boundary condition is a fundamental type of boundary condition
    extensively employed in Finite Element Analysis (FEA) to define precise
    values or limitations along certain edges or surfaces within a computational
    domain.

    Mathematically, the Dirichlet boundary condition is utilized to enforce or
    set the solution (such as displacement) at specific boundary points. This is
    done to accurately model real-world restrictions or established behaviors in
    the context of FEA. 

    Attributes:
        displacement_x (float): The imposed displacement in x-direction. Only
          fill it if want to resprtit the displamente. 
          Fill this attribute if you want to specify the displacement.
        displacement_y (float): The imposed displacement in y-direction. Only
          fill it if want to resprtit the displamente.
          Fill this attribute if you want to specify the displacement.
    """

    def __init__(self,
                 boundary_name: Literal["left", "top", "right", "bottom"],
                 displacement_x: Optional[float] = None,
                 displacement_y: Optional[float] = None) -> None:
        """Initializes a DirichletBC object."""
        super().__init__(boundary_name)
        self.displacement_x = displacement_x
        self.displacement_y = displacement_y

    def to_dict(self) -> dict:
        """Convert hole properties to a dictionary.

        Returns:
            dict: Hole properties.
        """
        return {
            "boundary_name": self.boundary_name,
            "displacement_x": self.displacement_x,
            "displacement_y": self.displacement_y
        }


class NeumannBC(BoundaryCondition):
    """A Neumann boundary condition is a vital aspect of Finite Element Analysis
    (FEA), used to specify the flux or the rate of change of a solution variable
    across certain boundaries within a computational domain.

    In mathematical terms, the Neumann boundary condition is employed to
    describe the gradient, flux, or external influence applied to the solution 
    (such as stress) along specific boundary regions. This type of boundary 
    condition is highly valuable in FEA when precise information is available
    about external forces. 

    Attributes:
        tension_x (float): The tension value in x-direction.
        tension_y (float): The tension value in y-direction.
    """

    def __init__(self, boundary_name: Literal["left", "top", "right", "bottom"],
                 tension_x: float, tension_y: float) -> None:
        """Initializes a NeumannBC object."""
        super().__init__(boundary_name)
        self.tension_x = tension_x
        self.tension_y = tension_y

    def to_dict(self) -> dict:
        """Convert hole properties to a dictionary.

        Returns:
            dict: Hole properties.
        """
        return {
            "boundary_name": self.boundary_name,
            "tension_x": self.tension_x,
            "tension_y": self.tension_y
        }



"""Post-processing and visualization utilities of the coastal area scenario."""

import os
import tempfile
from typing import Dict, Literal, Optional, Sequence, Tuple

import matplotlib.pyplot as plt
import numpy as np
import scipy

from inductiva.types import Path
from inductiva.utils import visualization

# Labels used in SWASH simulation outputs.
QUANTITY_SWASH_LABELS = {
    "water_level": "WATLEV",
    "velocity_x": "VKSI",
    "velocity_y": "VETA",
    "velocity_magnitude": "VMAG",
}

# Labels used in plots.
QUANTITY_PLOT_LABELS = {
    "water_level": "Water level",
    "velocity_x": "Velocity x-component",
    "velocity_y": "Velocity y-component",
    "velocity_magnitude": "Velocity magnitude",
}

# Units used in plots.
QUANTITY_UNITS = {
    "water_level": "m",
    "velocity_x": "m/s",
    "velocity_y": "m/s",
    "velocity_magnitude": "m/s",
}

GRID_POSITIONS_FILE_NAME = "grid_positions.mat"


def read_swash_output_file(file_path: str) -> Dict[str, np.ndarray]:
    """Reads a SWASH output file.
    
    The output is read as a dictionary, where the keys are the names of the
    variables and the values are the corresponding arrays.
    """

    try:
        # Read file contents
        data_dict = scipy.io.loadmat(file_path)

    except:
        raise ValueError(f"Error reading file '{file_path}'.") from None

    return data_dict


def read_swash_grid_positions_file(file_path: str) -> Tuple[np.ndarray]:
    """Reads a SWASH output file containing grid positions."""

    # Read grid positions
    grid_positions_dict = read_swash_output_file(file_path)

    # Get grids with x and y coordinates
    x_grid = grid_positions_dict["Xp"].transpose()
    y_grid = grid_positions_dict["Yp"].transpose()

    # Get arrays with x and y coordinates
    x_array = x_grid[:, 0]
    y_array = y_grid[0, :]

    return x_array, y_array


def read_swash_grid_quantity_file(file_path: str, quantity: str):
    """Reads a SWASH output file containing a quantity defined over a grid.

    SWASH outputs are stored in .mat files. Here, we first read the contents
    of one of these files as a dictionary, and then convert the dictionary
    to a list of times and a list of data arrays.

    # The list of times is built from the keys of the dictionary:
    # the keys end with a number, in format 123456_789, identifying the time
    # instant 12 hours, 34 minutes, 56 seconds and 789 milliseconds; we
    # strip the times from the keys and convert them to seconds.

    # The list of data arrays is built from the values of the dictionary:
    # the values correspond to the data values at each of the time instants
    # represented by the corresponding keys; we read each value as a 2D
    # tensor with indices (x, y).
    """

    # Read data file
    data_dict = read_swash_output_file(file_path)

    # Convert data dictionary to lists of times and data arrays
    time_list = []
    data_list = []

    for key, data_values in data_dict.items():
        # Check if key starts with the quantity label attributed by SWASH
        if key.startswith(QUANTITY_SWASH_LABELS[quantity].capitalize()):
            # Parse key string to get time in milliseconds
            time_ms = key[len(QUANTITY_SWASH_LABELS[quantity]):].replace(
                "_", "")

            # Convert time to seconds
            time_s = float(time_ms[:2]) * 3600. + \
                     float(time_ms[2:4]) * 60. + \
                     float(time_ms[4:6]) + \
                     float(time_ms[6:]) / 1000.

            # The values are transposed in the file, so we transpose them
            # again for consistency
            transposed_data_values = data_values.transpose()

            time_list.append(time_s)
            data_list.append(transposed_data_values)

    return time_list, data_list


def _render_quantity_grid_data(x_array: np.ndarray,
                               y_array: np.ndarray,
                               time_list: Sequence[float],
                               data_list: Sequence[np.ndarray],
                               quantity: str,
                               movie_path: str,
                               fps: int = 10,
                               cmap: Optional[str] = None,
                               clim: Optional[Sequence[float]] = None):
    """Renders temporal evolution of a quantity defined over a grid."""

    with tempfile.TemporaryDirectory() as tmp_dir:

        axes_limits = [
            x_array.min(),
            x_array.max(),
            y_array.min(),
            y_array.max(),
        ]

        if clim is None:
            clim = [
                np.min([np.min(data) for data in data_list]),
                np.max([np.max(data) for data in data_list]),
            ]

        for idx, (time, data) in enumerate(zip(time_list, data_list)):
            fig = plt.figure()
            ax = fig.add_subplot()

            im = ax.imshow(
                np.transpose(data),
                origin="lower",
                cmap=cmap,
                clim=clim,
                extent=axes_limits,
                aspect="equal",
            )

            plt.colorbar(
                im,
                ax=ax,
                label=(f"{QUANTITY_PLOT_LABELS[quantity]} "
                       f"[{QUANTITY_UNITS[quantity]}]"),
            )

            ax.set(
                xlabel="x [m]",
                ylabel="y [m]",
                title=f"Time: {time:.2f} [s]",
            )

            fig.savefig(os.path.join(tmp_dir, f"{idx:06d}.png"))
            plt.close(fig)

        visualization.create_movie_from_frames(frames_dir=tmp_dir,
                                               movie_path=movie_path,
                                               fps=fps)


class CoastalAreaOutput:
    """Heat sink simulation output."""

    def __init__(self, sim_output_path: Path):
        """Initializes a `HeatSinkOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
        """

        self.sim_output_path = sim_output_path

    def render(
        self,
        quantity: Literal[
            "water_level",
            "velocity_x",
            "velocity_y",
            "velocity_magnitude",
        ] = "water_level",
        movie_path: Path = "movie.mp4",
        fps: int = 10,
        cmap: str = "viridis",
        clim: Optional[Sequence[float]] = None,
    ):
        """Renders temporal evolution of a physical quantity.
        
        A movie is produced representing the temporal evolution of a physical
        quantity exported in simulations of the coastal area scenario. Available
        quantities are: water_level, velocity_x, velocity_y, velocity_magnitude.
         
        The movie is saved in the `movie_path` location.

        Args:
            quantity: Quantity to render. Available quantities are: water_level,
              velocity_x, velocity_y, velocity_magnitude.
            movie_path: Path to the movie file.
            fps: Number of frames per second to use in the movie.
            cmap: Colormap used to represent the quantity.
            clim: Colorbar limits. When not specified, the limits are
              automatically determined from the data.
        """

        x_array, y_array = read_swash_grid_positions_file(
            os.path.join(self.sim_output_path, GRID_POSITIONS_FILE_NAME))

        time_list, data_list = read_swash_grid_quantity_file(
            file_path=os.path.join(self.sim_output_path, quantity + ".mat"),
            quantity=quantity,
        )

        _render_quantity_grid_data(
            x_array=x_array,
            y_array=y_array,
            time_list=time_list,
            data_list=data_list,
            quantity=quantity,
            movie_path=movie_path,
            fps=fps,
            cmap=cmap,
            clim=clim,
        )



"""Methods to interact with the user storage resources."""
from absl import logging
import inductiva
from inductiva.client.apis.tags import instance_api


def get_space_used():
    """Returns the occupied storage size in GB."""
    try:
        api = instance_api.InstanceApi(inductiva.api.get_client())
        instance_price = api.get_storage_size()
        logging.info("Total storage used: %s GB",
                     float(round(instance_price.body, 3)))
        return float(round(instance_price.body, 3))
    except inductiva.client.ApiException as api_exception:
        raise api_exception



"""MachineGroup class to manage Google Cloud resources."""
import typing
import time

from absl import logging

import inductiva
from inductiva import api
from inductiva.client.apis.tags import instance_api


class MachineGroup():
    """Class to manage Google Cloud resources."""

    def __init__(
        self,
        machine_type: str,
        num_machines: int = 1,
        spot: bool = False,
        disk_size_gb: int = 30,
        zone: typing.Optional[str] = "europe-west1-b",
    ) -> None:
        """Create a MachineGroup object.

        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-medium".
            Check https://cloud.google.com/compute/docs/machine-resource for
            more information about machine types.
            num_machines: The number of virtual machines to launch.
            spot: Whether to use spot machines.
            disk_size_gb: The size of the disk in GB, recommended min. is 30 GB.
            zone: The zone where the machines will be launched.
        """
        self.id = None
        self.name = None
        #TODO: Check if machine type is valid.
        self.machine_type = machine_type
        self.num_machines = num_machines
        self.spot = spot
        self.disk_size_gb = disk_size_gb
        self.zone = zone

        # Set the API configuration that carries the information from the client
        # to the backend.
        self._api = instance_api.InstanceApi(api.get_client())

    def start(self):
        """Starts a machine group."""
        instance_group_config = \
            inductiva.client.model.instance_group.InstanceGroup(
                machine_type=self.machine_type,
                num_instances=self.num_machines,
                spot=self.spot,
                disk_size_gb=self.disk_size_gb,
                zone=self.zone,
            )
        try:
            logging.info("Creating a machine group."
                         "This may take up to a few minutes.")
            start_time = time.time()
            instance_group = self._api.create_instance_group(
                body=instance_group_config)
            creation_time_mins = (time.time() - start_time) / 60

            self.id = instance_group.body["id"]
            self.name = instance_group.body["name"]
            self.estimated_price = self.estimate_price()

            logging.info("Machine group successfully created in %.2f mins.",
                         creation_time_mins)
            self._log_machine_group_info()

        except inductiva.client.ApiException as api_exception:
            raise api_exception

    def terminate(self):
        """Terminates a machine group."""

        try:
            logging.info("Terminating machine group."
                         "This may take up to a few minutes.")
            start_time = time.time()

            instance_group_config = \
                inductiva.client.model.instance_group.InstanceGroup(
                    name=self.name,
                    machine_type=self.machine_type,
                    num_instances=self.num_machines,
                    spot=self.spot,
                    disk_size_gb=self.disk_size_gb,
                    zone=self.zone)
            self._api.delete_instance_group(body=instance_group_config)
            termination_time_mins = (time.time() - start_time) / 60
            logging.info(
                "Machine group of %s machines successfully"
                "terminated in %.2f mins.", self.num_machines,
                termination_time_mins)

        except inductiva.client.ApiException as api_exception:
            raise api_exception

    def estimate_price(self):
        """Returns an estimated price per hour of a machine group."""
        #TODO: Contemplate disk size in the price.
        instance_price = self._api.get_instance_price({
            "machine_type": self.machine_type,
            "zone": self.zone,
        })
        if self.spot:
            estimated_price = instance_price.body[
                "preemptible_price"] * self.num_machines
        else:
            estimated_price = instance_price.body[
                "on_demand_price"] * self.num_machines
        estimated_price = float(round(estimated_price, 3))
        logging.info("Estimated price per hour: %s $/h", estimated_price)

        return estimated_price

    def status(self):
        """Returns the status of a machine group if it exists.

        Otherwise returns None"""
        response = self._api.get_group_status({"name": self.name})

        if response.body == "notFound":
            logging.info("Machine group does not exist: %s.", self.name)
        return response.body

    def _log_machine_group_info(self):
        """Logs the machine group info."""

        logging.info("ID: %s", self.id)
        logging.info("Machine type: %s", self.machine_type)
        logging.info("Number of machines: %s", self.num_machines)
        logging.info("Spot: %s", self.spot)
        logging.info("Disk size: %s GB", self.disk_size_gb)
        logging.info("Estimated cost per hour: %s $/h", self.estimated_price)



#pylint: disable=missing-module-docstring
from .scenarios import DeformablePlate
from .simulators import FEniCSx
from . import bcs
from . import holes
from . import materials
from . import plates
from . import scenarios



"Postprocessing steps for the MDWaterBox scenario."
import os
import pathlib
import time
from typing import Literal

import matplotlib.pyplot as plt
import MDAnalysis as mda
import nglview as nv
import numpy as np

import inductiva

FULL_TRAJECTORY_FILE = "full_trajectory.trr"
COMPRESSED_TRAJECTORY_FILE = "compressed_trajectory.xtc"
PROTEIN_TOPOLOGY_FILE = "protein.gro"
SYSTEM_TOPOLOGY_FILE = "solvated_protein.tpr"


class ProteinSolvationOutput:
    """Post process the simulation output of a ProteinSolvation scenario."""

    def __init__(self, sim_output_path: pathlib.Path = None):
        """Initializes a `ProteinSolvationOutput` object.

        Given a simulation output directory that contains the standard files
        generated by a ProteinSolvation simulation run, this class provides 
        methods to visualize the simulation outputs in a notebook interactively.

        Args:
            sim_output_path: Path to the simulation output directory."""

        self.sim_output_dir = sim_output_path
        topology_path = os.path.join(self.sim_output_dir, PROTEIN_TOPOLOGY_FILE)
        trajectory_path = os.path.join(self.sim_output_dir,
                                       COMPRESSED_TRAJECTORY_FILE)
        self.universe = inductiva.molecules.scenarios.utils.unwrap_trajectory(
            topology_path, trajectory_path)

    def render_interactive(self,
                           representation: Literal["cartoon", "ball+stick",
                                                   "line", "point", "surface",
                                                   "ribbon"] = "ball+stick",
                           selection: str = "protein",
                           add_backbone: bool = True):
        """
        Render the simulation outputs in an interactive visualization.
        Args: 
            representation: The protein representation to use for the 
            visualization.
            selection: The selection to use for the representation. Check 
            https://nglviewer.org/ngl/api/manual/usage/selection-language.html 
            for details.
            add_backbone: Whether to add the protein backbone to the 
            visualization.
            """
        view = nv.show_mdanalysis(self.universe)
        view.add_representation(representation, selection=selection)
        if add_backbone:  #hardcoding the backbone as a cartoon representation
            view.add_representation("cartoon", selection="protein")
        view.center()

        print("System Information: ")
        print(f"Number of protein atoms: {len(self.universe.atoms)}")
        print(f"Number of amino acids: "
              f"{self.universe.select_atoms('protein').n_residues}")
        print(f"Number of trajectory frames: {len(self.universe.trajectory)}")
        return view

    def calculate_rmsf_trajectory(self):
        """Calculate the root mean square fluctuation (RMSF) over a trajectory.  

        It is typically calculated for the alpha carbon atom of each residue. 
        These atoms make the backbone of the protein.The RMSF is the square root 
        of the variance of the fluctuation around the average position:
        &rhoi = √⟨(xi - ⟨xi⟩)²⟩
        It quantifies how much a structure diverges from the average structure 
        over time, the RSMF can reveal which areas of the system are the most 
        mobile. Check 
        https://userguide.mdanalysis.org/stable/examples/analysis/alignment_and_rms/rmsf.html 
        for more details.

        Args:
            nglview_visualization: Whether to return visualization of the 
            RMSF using nglview or not."""
        start_time = time.time()
        topology_path = os.path.join(self.sim_output_dir, SYSTEM_TOPOLOGY_FILE)
        full_trajectory_path = os.path.join(self.sim_output_dir,
                                            FULL_TRAJECTORY_FILE)
        full_precision_universe = mda.Universe(topology_path,
                                               full_trajectory_path)

        aligned_trajectory_path = os.path.join(self.sim_output_dir,
                                               "aligned_traj.dcd")
        inductiva.molecules.scenarios.utils.align_trajectory_to_average(
            full_precision_universe, aligned_trajectory_path)
        align_universe = mda.Universe(topology_path, aligned_trajectory_path)

        # Calculate RMSF for carbon alpha atoms
        c_alphas = align_universe.select_atoms("protein and name CA")
        rmsf = mda.analysis.rms.RMSF(c_alphas).run()
        rmsf_values = rmsf.results.rmsf
        duration = time.time() - start_time
        print(f"RMSF calculation took {duration:.2f} seconds.")
        return rmsf_values

    def plot_rmsf_per_residue(self, rmsf_values: np.array):
        "Plot RMSF values per residue."
        # Plot the data
        residue_number = np.arange(len(rmsf_values))
        plt.plot(residue_number, rmsf_values)
        plt.xlabel("Residue Number")
        plt.ylabel("RMSF")
        plt.title("RMSF per residue")
        plt.grid(True)
        plt.show()

    def render_attribute_per_residue(
        self,
        residue_attributes: np.ndarray,
        representation: Literal["cartoon", "ball+stick", "line", "point",
                                "surface", "ribbon"] = "cartoon"):
        """Render a specific protein attribute in an interactive visualization.
        Args: 
            residue_attributes: The per residue values of the attribute you want 
            to visualize.
            representation: The protein representation to use for the 
            visualization.
            """
        self.universe.add_TopologyAttr("tempfactors")
        protein = self.universe.select_atoms("protein")
        for residue, value in zip(protein.residues, residue_attributes):
            residue.atoms.tempfactors = value
        view = nv.show_mdanalysis(self.universe)
        view.add_representation(representation, selection="protein")
        view.update_representation(color_scheme="bfactor")
        view.center()
        return view



"""DualSPHysics module of the API."""
from typing import Literal, Optional

from inductiva import types, tasks, resources
from inductiva.simulation import Simulator


class SPlisHSPlasH(Simulator):
    """Class to invoke a generic SPlisHSPlasH simulation on the API."""

    @property
    def api_method_name(self) -> str:
        return "sph.splishsplash.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        machine_group: Optional[resources.MachineGroup] = None,
        device: Literal["gpu", "cpu"] = "cpu",
        run_async: bool = False,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            sim_config_filename: Name of the simulation configuration file.
            device: Device in which to run the simulation.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            machine_group=machine_group,
            device=device,
            input_filename=sim_config_filename,
            run_async=run_async,
        )



"""Post-processing tools of fluid dynamics steady-state simulations.

This class implements various post-processing capabilities for
the visualizations associated. Namely:
    - Pressure over object;
    - Cutting plane;
    - Stream lines.

Current Support - OpenFOAM
"""
import os
from dataclasses import dataclass
from enum import Enum
from typing import Literal
import pathlib
import glob
from absl import logging

import pyvista as pv

from inductiva import types, utils


class SteadyStateOutput:
    """Post-Process steady-state simulation outputs.

    This class contains several methods to post-process the output 
    and visualize the results of a steady-state simulation where
    time-independent results are obtained.

    To be general we assume that a simulation is performed over a
    regular box - the domain - and a certain object placed inside.
    The object is either a small object inside the domain or a more
    general object that spreads all through the domain.
    """

    def __init__(self, sim_output_path: types.Path):
        """Initializes a `SteadyStateOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
            default_output_files_list: List of default output files
                that are computed on the backend and saved to the
                simulation output directory.
        """

        self.sim_output_path = sim_output_path
        self.full_output = self.inspect_output()

    def inspect_output(self):
        """Inspect the output of the simulation.
        
        Return True if the simulation output contains more files than
        the default ones. This is used to determine if the post-processing
        was computed on the backend or not."""

        # TODO: Transport these files in a simpler way from the task.
        # For such, a change to all of the post-processing output classes may
        # be required.
        default_output_files_list = [
            "pressure_field.vtk", "streamlines.vtk", "stdout.txt", "stderr.txt",
            "force_coefficients.csv", "xy_flow_slice.vtk", "xz_flow_slice.vtk",
            "yz_flow_slice.vtk", "object.obj"
        ]

        sim_output_files = glob.glob(os.path.join(self.sim_output_path, "**",
                                                  "*.*"),
                                     recursive=True)

        for index, file in enumerate(sim_output_files):
            sim_output_files[index] = file.split("/")[-1]

        return sorted(sim_output_files) != sorted(default_output_files_list)

    def get_last_iteration(self):
        """Get the last iteration of the simulation from simulation outputs."""

        # Sort all output folders and files by alphabetical order.
        outputs_dir_list = sorted(os.listdir(self.sim_output_path))
        # Second folder contains the last iteration outputs.
        last_iteration = float(outputs_dir_list[1])

        return last_iteration

    def get_output_mesh(self):
        """Get domain and object mesh info at the steady-state."""

        if not self.full_output:
            object_mesh = pv.read(
                os.path.join(self.sim_output_path, "constant", "triSurface",
                             "object.obj"))
            return object_mesh

        # The OpenFOAM data reader from PyVista requires that a file named
        # "foam.foam" exists in the simulation output directory.
        # Create this file if it does not exist.
        foam_file_path = os.path.join(self.sim_output_path, "foam.foam")
        pathlib.Path(foam_file_path).touch(exist_ok=True)

        reader = pv.OpenFOAMReader(foam_file_path)
        reader.set_active_time_value(self.get_last_iteration())

        full_mesh = reader.read()
        domain_mesh = full_mesh["internalMesh"]
        object_mesh = full_mesh["boundary"]["object"]

        return domain_mesh, object_mesh

    def get_object_pressure_field(self, save_path: types.Path = None):
        """Get a pressure scalar field over mesh points of the object.

        Returns:
            A MeshData object that allow to manipulate the data over a mesh
            and to render it.
        """

        if not self.full_output:
            logging.info("Pressure field was computed on the backend. "
                         "Args passed here were ignored.")
            object_mesh = pv.read(
                os.path.join(self.sim_output_path, "pressure_field.vtk"))
        else:
            logging.info("Fetching pressure field over the object.")
            _, object_mesh = self.get_output_mesh()

        field_notation = OpenFOAMPhysicalField["PRESSURE"].value
        pressure_field = MeshData(object_mesh, field_notation)

        if save_path is not None:
            save_path = utils.files.resolve_path(save_path)
            pressure_field.mesh.save(save_path)

        return pressure_field

    def get_streamlines(self,
                        max_time: float = 100,
                        n_points: int = 100,
                        initial_step_length: float = 1,
                        source_radius: float = 0.7,
                        source_center: tuple = (0, 0, 0),
                        save_path: types.Path = None):
        """Get streamlines through the fluid/domain.
        
        The streamlines are obtained by seeding a set of points
        at the inlet of the WindTunnel.

        Args:
            simulation_time: Time value to obtain simulation mesh.
            max_time: Time used for integration of the streamlines.
                Not related with simulation time.
            n_points: Number of points to seed.
            initial_step_length: Initial step length for the streamlines.
            source_radius: Radius of the source of the streamlines.
            source_center: Center of the source of the streamlines.
            save_path: Path to save the streamlines. 
                Types of files permitted: .vtk, .ply, .stl
        """

        if not self.full_output:
            logging.info("Streamlines were computed on the backend. "
                         "Args passed here were ignored.")
            object_mesh = self.get_output_mesh()
            streamlines_mesh = pv.read(
                os.path.join(self.sim_output_path, "streamlines.vtk"))
        else:
            logging.info("Computing the streamlines on the fly.")
            mesh, object_mesh = self.get_output_mesh()

            streamlines_mesh = mesh.streamlines(
                max_time=max_time,
                n_points=n_points,
                initial_step_length=initial_step_length,
                source_radius=source_radius,
                source_center=source_center)

            if save_path is not None:
                save_path = utils.files.resolve_path(save_path)
                streamlines_mesh.save(save_path)

        return Streamlines(object_mesh, streamlines_mesh)

    def get_flow_slice(self,
                       plane: Literal["xy", "xz", "yz"] = "xz",
                       origin: tuple = (0, 0, 0),
                       save_path: types.Path = None):
        """Get flow properties in a slice of the domain.
        
        Args:
            simulation_time: Time value to obtain simulation mesh.
            plane: Orientation of the plane to slice the domain.
            origin: Origin of the plane.
            save_path: Path to save the flow slice. 
                Types of files permitted: .vtk, .ply, .stl
        """

        if not self.full_output:
            logging.info("Flow slices were computed on the backend. "
                         "The origin and save_path arg are ignored.")
            object_mesh = self.get_output_mesh()
            flow_slice_name = plane + "_flow_slice.vtk"
            flow_slice = pv.read(
                os.path.join(self.sim_output_path, flow_slice_name))
        else:
            logging.info("Computing the flow on a slice.")
            mesh, object_mesh = self.get_output_mesh()

            if plane == "xy":
                normal = (0, 0, 1)
            elif plane == "yz":
                normal = (1, 0, 0)
            elif plane == "xz":
                normal = (0, 1, 0)
            else:
                raise ValueError("Invalid view.")

            flow_slice = mesh.slice(normal=normal, origin=origin)

            if save_path is not None:
                save_path = utils.files.resolve_path(save_path)
                flow_slice.save(save_path)

        return FlowSlice(object_mesh, flow_slice)


@dataclass
class OpenFOAMPhysicalField(Enum):
    """Defines the notation used for physical field in OpenFOAM."""
    PRESSURE = "p"
    VELOCITY = "U"


class FlowSlice:
    """Render flow field in a plane of the domain."""

    def __init__(self, object_mesh, flow_slice):
        self.object_mesh = object_mesh
        self.mesh = flow_slice

    def render(self,
               physical_field: Literal["pressure", "velocity"] = "pressure",
               off_screen: bool = False,
               virtual_display: bool = False,
               background_color: str = "white",
               flow_cmap: str = "viridis",
               object_color: str = "white",
               save_path: types.Path = None):
        """Render flow property over domain."""

        if save_path is not None:
            save_path = utils.files.resolve_path(save_path)

        if virtual_display:
            pv.start_xvfb()

        plotter = pv.Plotter(off_screen=off_screen)
        plotter.background_color = background_color

        center = self.mesh.center
        normal = -self.mesh.compute_normals()["Normals"].mean(axis=0)
        plotter.camera.position = center + normal
        plotter.camera.focal_point = center
        plotter.camera.zoom(1.2)

        # Obtain notation for the physical field for the simulator.
        field_notation = OpenFOAMPhysicalField[physical_field.upper()].value

        plotter.add_mesh(self.object_mesh, color=object_color)
        plotter.add_mesh(self.mesh, scalars=field_notation, cmap=flow_cmap)
        plotter.reset_camera(bounds=self.mesh.bounds)
        plotter.show(screenshot=save_path)
        plotter.close()


class Streamlines:
    """Class to render streamlines through the domain with object."""

    def __init__(self, object_mesh, streamlines):
        self.object_mesh = object_mesh
        self.mesh = streamlines

    def render(self,
               physical_field: Literal["pressure", "velocity"] = "pressure",
               off_screen: bool = False,
               virtual_display: bool = False,
               background_color: str = "white",
               flow_cmap: str = "viridis",
               view: Literal["isometric", "front", "rear", "top",
                             "side"] = "isometric",
               object_color: str = "white",
               streamline_radius: float = 0.1,
               save_path: types.Path = None):
        """Render streamlines through domain."""

        if save_path is not None:
            save_path = utils.files.resolve_path(save_path)

        if virtual_display:
            pv.start_xvfb()

        plotter = pv.Plotter(off_screen=off_screen)
        plotter.background_color = background_color

        # Set camera position for a nice view.
        if view == "isometric":
            plotter.view_vector([-1, -2, 1], viewup=[0.31, 0.90, 0.29])
        elif view == "front":
            plotter.view_yz(negative=True)
        elif view == "rear":
            plotter.view_yz()
        elif view == "top":
            plotter.view_xy()
        elif view == "side":
            plotter.view_xz()
        else:
            raise ValueError("Invalid view.")

        # Obtain notation for the physical field for the simulator.
        field_notation = OpenFOAMPhysicalField[physical_field.upper()].value

        plotter.add_mesh(self.mesh.tube(radius=streamline_radius),
                         scalars=field_notation,
                         cmap=flow_cmap)

        plotter.add_mesh(self.object_mesh, color=object_color)
        # Slide along the vectord defined from camera position to focal point,
        # until all of the meshes are visible.
        plotter.reset_camera(bounds=self.mesh.bounds, render=False)
        plotter.show(screenshot=save_path)
        plotter.close()


class MeshData:
    """MeshData class that allows for mesh data manipulation and render.

    Process outputs of simulation over meshes. We assume that scalar
    fields are defined over the points of the mesh. In this way, we have
    a general method to manipulate the data and render the specification
    data. As of now, this serves a niche, but the goal
    is to integrate this more generally later on.

    Example:
    Imagine we run a WindTunnel simulation with an object inside
    for which a mesh was generated. A possible output of the simulation is
    a mesh of the object with scalar fields over the points defining a certain
    physical property, e.g., pressure.

    Assume this output mesh is `object_mesh`. To process the pressure field over
    the data, we can do `pressure_field = MeshData(object_mesh, "p")`

    `pressure_field.mesh` contains a mesh structure with the pressure field over
    the mesh.

    `pressure_field.render()` renders the pressure field property over the mesh.
    """

    def __init__(self, mesh_data, scalar_name: str = None):
        """Initialize a `MeshData` type object.

        Args:
            mesh_data: pyvista mesh (PolyData or Unstructured) which
                contains all of the simulated outputs over the mesh.
            scalar_name: string that defines the scalar we want to
                manipulate and render. This names depends on the mesh_data
                array names and on the specific simulator used.

        Attributes:
            mesh: Mesh over the object obtained from the simulation.
        """
        self.mesh = pv.PolyData(mesh_data.points, faces=mesh_data.faces)
        self.scalar_name = scalar_name
        self.mesh.point_data[scalar_name] = mesh_data.point_data[scalar_name]
        self.mesh.cell_data[scalar_name] = mesh_data.cell_data[scalar_name]

    def render(self,
               off_screen: bool = False,
               background_color: str = "black",
               scalars_cmap: str = "viridis",
               virtual_display: bool = False,
               save_path: types.Path = None):
        """Render scalar field data over the mesh."""
        if save_path is not None:
            save_path = utils.files.resolve_path(save_path)

        if virtual_display:
            pv.start_xvfb()

        plotter = pv.Plotter(off_screen=off_screen)
        pv.global_theme.background = background_color
        plotter.add_mesh(self.mesh, scalars=self.scalar_name, cmap=scalars_cmap)
        plotter.show(screenshot=save_path)
        plotter.close()



"""FEniCSx module of the API for Finite Element Analysis."""

from typing import Optional

import inductiva


class FEniCSx(inductiva.simulation.Simulator):
    """Class to invoke a generic FEniCSx simulation on the API."""

    def __init__(self, api_method: str = "fem.fenicsx.run_simulation"):
        super().__init__()
        self.api_method = api_method

    @property
    def api_method_name(self) -> str:
        return self.api_method

    def run(
        self,
        input_dir: inductiva.types.Path,
        mesh_filename: str,
        bcs_filename: str,
        material_filename: str,
        machine_group: Optional[inductiva.resources.MachineGroup] = None,
        run_async: bool = False,
    ) -> inductiva.tasks.Task:
        """Run the simulation.

        Args:
            mesh_filename: Mesh filename.
            bcs_filename: Boundary conditions filename.
            material_filename: Material filename.
            machine_group: The machine group to use for the simulation.
            run_async: Whether to run the simulation asynchronously.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           machine_group=machine_group,
                           run_async=run_async,
                           mesh_filename=mesh_filename,
                           bcs_filename=bcs_filename,
                           material_filename=material_filename)



"""Protein solvation scenario."""

from functools import singledispatchmethod
from typing import Optional, Literal
import os
import shutil

from inductiva import tasks, resources
from inductiva.molecules.simulators import GROMACS
from inductiva.simulation import Simulator
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       batch_replace_params_in_template)
from inductiva.scenarios import Scenario
from inductiva.utils import files
from .post_processing import ProteinSolvationOutput

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "protein_solvation")
GROMACS_TEMPLATE_INPUT_DIR = "gromacs"
COMMANDS_TEMPLATE_FILE_NAME = "commands.json"


class ProteinSolvation(Scenario):
    """Solvated protein scenario."""

    valid_simulators = [GROMACS]

    def __init__(self, protein_pdb: str, temperature: float = 300):
        """
        Scenario constructor for protein solvation based on the GROMACS
        simulator.
        The three main steps of this scenario are solvation, energy minimization
        and simulation. The user can control the number of steps used to perform
        the energy minimization step and the duration, temperature and
        integrator used to perform the simulation.
        Args:
            protein_pdb: The path to the protein pdb file.
            temperature: The temperature to use for the simulation.
            charged: Whether the protein is charged or not. If None, the charge
            is computed automatically.
        """
        self.template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                         GROMACS_TEMPLATE_INPUT_DIR)
        self.protein_pdb = files.resolve_path(protein_pdb)
        self.temperature = temperature

    def simulate(
            self,
            simulator: Simulator = GROMACS("proteinsolvation"),
            machine_group: Optional[resources.MachineGroup] = None,
            run_async: bool = False,
            simulation_time_ns: float = 10,  # ns
            output_timestep_ps: float = 1,  # ps
            integrator: Literal["md", "sd", "bd"] = "md",
            n_steps_min: int = 5000) -> tasks.Task:
        """Simulate the solvation of a protein.

        Args:
            machine_group: The machine group to use for the simulation.
            simulation_time_ns: The simulation time in ns.
            output_timestep_ps: The output timestep in ps.
            integrator: The integrator to use for the simulation. Options:
                - "md" (Molecular Dynamics): Accurate leap-frog algorithm for
                integrating Newton's equations of motion.
                - "sd" (Steepest Descent): Stochastic dynamics integrator with
                leap-frog scheme.
                - "bd" (Brownian Dynamics): Euler integrator for Brownian or
                position Langevin dynamics.

            For more details on the integrators, refer to the GROMACS
            documentation at
            https://manual.gromacs.org/current/user-guide/mdp-options.html.

            n_steps_min: Number of steps for energy minimization.
            run_async: Whether to run the simulation asynchronously.
        """

        self.nsteps = int(
            simulation_time_ns * 1e6 / 2
        )  # convert to fs and divide by the time step of the simulation (2 fs)
        self.output_frequency = int(
            output_timestep_ps * 1000 /
            2)  # convert to fs and divide by the time step
        # of the simulation (2 fs)
        self.integrator = integrator
        self.n_steps_min = n_steps_min
        commands = self.get_commands()
        task = super().simulate(simulator,
                                machine_group=machine_group,
                                commands=commands,
                                run_async=run_async)

        task.set_output_class(ProteinSolvationOutput)

        return task

    def get_commands(self):
        """Returns the commands for the simulation."""

        commands_template_path = os.path.join(SCENARIO_TEMPLATE_DIR,
                                              GROMACS_TEMPLATE_INPUT_DIR,
                                              COMMANDS_TEMPLATE_FILE_NAME)

        commands = self.read_commands_from_file(commands_template_path)
        return commands

    @singledispatchmethod
    def create_input_files(self, simulator: Simulator):
        pass


@ProteinSolvation.create_input_files.register
def _(self, simulator: GROMACS, input_dir):  # pylint: disable=unused-argument
    """Creates GROMACS simulation input files."""

    # rename the pdb file to comply with the naming in the commands list
    shutil.copy(self.protein_pdb, os.path.join(input_dir, "protein.pdb"))

    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      GROMACS_TEMPLATE_INPUT_DIR)

    shutil.copytree(template_files_dir, input_dir, dirs_exist_ok=True)

    batch_replace_params_in_template(
        templates_dir=input_dir,
        template_filenames=[
            "simulation.mdp.jinja",
            "energy_minimization.mdp.jinja",
        ],
        params={
            "integrator": self.integrator,
            "nsteps": self.nsteps,
            "ref_temp": self.temperature,
            "nsteps_minim": self.n_steps_min,
            "output_frequency": self.output_frequency,
        },
        output_filename_paths=[
            os.path.join(input_dir, "simulation.mdp"),
            os.path.join(input_dir, "energy_minimization.mdp"),
        ],
        remove_templates=True,
    )



"""Utils to create plates."""

from abc import ABC, abstractmethod
from typing import Tuple

import gmsh


class Plate(ABC):
    """Abstract base class for plates."""

    @abstractmethod
    def perimeter(self):
        """Abstract method to calculate the perimeter of the plate."""
        pass

    @abstractmethod
    def to_dict(self):
        """Abstract method to convert the plate properties to a dictionary."""
        pass

    @abstractmethod
    def to_occ(self):
        """Abstract method to convert the plate to OpenCASCADE CAD."""
        pass

    @abstractmethod
    def get_plate_mesh_params(self):
        """Abstract method to get plate mesh parameters."""
        pass


class RectangularPlate(Plate):
    """Rectangular plate.

    Attributes:
        plate_type (str): Plate type.
        width (float): Plate width.
        length (float): Plate length.
    """

    def __init__(self, width: float, length: float) -> None:
        """Initializes a RectangularPlate object."""
        self.plate_type = "rectangular"
        self.width = width
        self.length = length

    def perimeter(self) -> float:
        """Calculate the perimeter of the plate.

        Returns:
            float: The calculated perimeter.
        """
        return 2 * (self.width + self.length)

    def to_dict(self) -> dict:
        """Convert plate properties to a dictionary.

        Returns:
            dict: Plate properties.
        """
        return {
            "plate_type": "rectangular",
            "length": self.length,
            "width": self.width
        }

    def to_occ(self) -> int:
        """Converts plate to OpenCASCADE CAD representation.

        Returns:
          plate_gmsh (int): The Gmsh entity ID representing the plate's
            OpenCASCADE CAD representation.
        """
        plate_gmsh = gmsh.model.occ.addRectangle(x=0,
                                                 y=0,
                                                 z=0,
                                                 dx=self.width,
                                                 dy=self.length)

        return plate_gmsh

    def get_plate_mesh_params(self) -> Tuple[float, float]:
        """Gets the mesh parameters for the plate.

        Metrics:
          - mesh_offset (float): Represents an offset for all the boundaries of
            the plate, defining a region around the boundaries. Within this
            region, we have the ability to control the mesh elements size.
            The offset is equal to half of the minimum size of the plate.
          - predefined_element_size (float): Represents the predefined element
            size, defined as 1/4 of the perimeter.

        Returns:
          Tuple[float, float]: A tuple containing the mesh offset and the
            predefined element mesh size for the plate.
      """
        mesh_offset = min(self.width, self.length) / 2
        predefined_element_size = self.perimeter() / 4

        return mesh_offset, predefined_element_size



"""Simsopt module of the API."""
from typing import Optional

from inductiva import simulation, tasks, types, resources


class Simsopt(simulation.Simulator):
    """Invokes a simsopt simulation on the API."""

    @property
    def api_method_name(self) -> str:
        return "stellarators.simsopt.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        plasma_surface_filename: str,
        coil_coefficients_filename: str,
        coil_currents_filename: str,
        num_field_periods: int,
        num_iterations: int,
        num_samples: int,
        sigma_scaling_factor: float,
        objectives_weights_filename: str,
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            coil_coefficients_filename: Name of the file with the Fourier
              Series coefficients of the coils.
            coil_currents_filename: Name of the file with the current in each
              coils.
            plasma_surface_filename: Name of the file with the description of
              the plasma surface on which the magnetic field will be calculated.
            num_field_periods: Number of magnetic field periods.
              Refers to the number of complete magnetic field repetitions 
              within a stellarator. Represents how many times the magnetic
              field pattern repeats itself along the toroidal direction.
            num_iterations: Number of iterations to run for the searching 
              process.
            num_samples: Number of different stellarator samples generated per 
              iteration from a configuration. The samples are generated using
              normal distribution noise for each coefficient of the Fourier 
              Series that describes the coils.
            sigma_scaling_factor: Scaling factor for the sigma value used in 
              the random generation of noise. This argument makes sure that 
              the noise is generated proportionally to each coefficient. It 
              also determines the range of search for new values of the
              coefficients.
            objectives_weights_filename: Name of the file with the weights for
              each objective function used in the construction of the total 
              objective.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            machine_group=machine_group,
            run_async=run_async,
            coil_coefficients_filename=coil_coefficients_filename,
            coil_currents_filename=coil_currents_filename,
            plasma_surface_filename=plasma_surface_filename,
            num_field_periods=num_field_periods,
            num_iterations=num_iterations,
            num_samples=num_samples,
            sigma_scaling_factor=sigma_scaling_factor,
            objectives_weights_filename=objectives_weights_filename)



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.paths.task_task_id_kill import Api

from inductiva.client.paths import PathValues

path = PathValues.TASK_TASK_ID_KILL